from hx_deep_learning_tools.dl_base_model import HxDeepLearningBaseModel
from _hx_model_evaluation_tools import EvaluateMulticlassClassifier, MulticlassClassifierMetricsCalculations
import tensorflow as tf
from _hx_model_evaluation_tools import DlShapToolsMulticlass
from typing import Dict, Literal, List, Any, Tuple
import pandas as pd
import numpy as np
import datetime
import os
import json
from sklearn.preprocessing import LabelEncoder


class HxDenseNeuralNetworkMulticlassClassifier(HxDeepLearningBaseModel):

    def __init__(self,
                 data_dict: Dict[str, pd.DataFrame],
                 hiperparams: dict,
                 metric: Literal['accuracy', 'precision_macro', 'recall_macro', 'f1_macro', 'f1_micro', 'balanced_accuracy', 'roc_auc_ovr'],
                 problem_type: Literal['multiclass'],
                 geometry: Literal[0, 1, 2, 3] = 1,
                 hidden_layers: int = 2,
                 activation: Literal['relu', 'tanh', 'sigmoid'] = 'relu',
                 regularization_type: Literal['l1', 'l2'] = 'l2',
                 optimizer: Literal['sgd', 'adam', 'rmsprop'] = 'adam',
                 include_dropout: bool = True,
                 include_batch_norm: bool = True,
                 early_stopping_patience: int = 15,
                 save_path: str | None = None,
                 bins: dict | None = None,
                 verbose: Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show',
                 random_state: int = 42):
        """
        Clase para construir modelos de clasificación multiclase con Redes Neuronales Densas
        :param data_dict: Diccionario que contiene el x_train, y_train, x_test, y_test
        :param save_path: Path donde guardar los resultados
        :param hiperparams: Hiperparámetros del modelo
        :param bins: Bins para discretización
        """
        super().__init__(data_dict,
                         hiperparams,
                         "DNN_multiclass_classifier",
                         geometry,
                         hidden_layers,
                         activation,
                         regularization_type,
                         optimizer,
                         include_dropout,
                         include_batch_norm,
                         early_stopping_patience,
                         save_path)

        # --------------------------------------------------------------------------------------------
        # -- 0: Almaceno nombre de la clase, pinto la entrada y creo la propiedad model
        # --------------------------------------------------------------------------------------------

        # ---- 0.1: Almaceno el nombre de la clase
        self.class_name: str = f"HxDenseNeuralNetworkMulticlassClassifier"

        # ---- 0.2: Pinto la entrada
        self.CT.IT.intro_print(self.class_name)

        # ---- 0.3: Creo la propiedad model y la inicializo en None
        self.model = None

        # ---- 0.4: Establecer semilla para reproducibilidad
        tf.random.set_seed(random_state)
        np.random.seed(random_state)

        # --------------------------------------------------------------------------------------------
        # -- 1: Asigno las propiedades abstractas
        # --------------------------------------------------------------------------------------------

        # ---- 1.1: Verbose
        self._verbose: Literal['show_all', 'show_basics', 'dont_show'] = verbose

        # ---- 1.2: Metric
        self._metric: Literal['accuracy', 'precision_macro', 'recall_macro', 'f1_macro', 'f1_micro', 'balanced_accuracy', 'roc_auc_ovr'] = metric

        # ---- 1.3: Problem type
        self._problem_type: Literal['binary', 'multiclass'] = problem_type

        # --------------------------------------------------------------------------------------------
        # -- 2: Asigno las propiedades estandar
        # --------------------------------------------------------------------------------------------

        # ---- 2.1: Si bins es None, pongo el bins por defecto, sino lo asigno
        if bins is None:
            self.bins: Dict[str, List] = {
                "bins": [0, 0.125, 0.25, 0.375, 0.5, 0.625, 0.75, 0.875, 1],
                "labels": ["0-12.5", "12.5-25", "25-37.5", "37.5-50", "50-62.5", "62.5-75", "75-87.5", "87.5-100"]
            }
        else:
            self.bins = bins

        # ---- 2.2: Path donde se va a guardar la informacion del modelo
        self.model_save_path: str = self.get_save_path()

        # ---- 2.3: Random state con el que se van a inicializar los pesos, por defecto 42
        self.random_state: int = random_state

        # ---- 2.4: Obtener número de clases y preparar los datos
        self.num_classes = len(np.unique(self.y_train))
        self.CT.IT.info_print(f"Número de clases detectadas: {self.num_classes}")

        # ---- 2.5: CONVERSIÓN CRÍTICA: Convertir y_train, y_test a one-hot encoding
        self._convert_targets_to_one_hot()

        # ---- 2.6: Creo los callbacks llamando al metodo de la clase padre
        self._setup_training_callbacks()

    def _convert_targets_to_one_hot(self):
        """
        Convierte las variables target (y_train, y_test) de labels enteros a one-hot encoding
        """
        # ---- 2.5.1: Aplanar y convertir a numpy array
        y_train_flat = self.y_train.values.ravel() if len(self.y_train.shape) > 1 else self.y_train
        y_test_flat = self.y_test.values.ravel() if len(self.y_test.shape) > 1 else self.y_test

        # ---- 2.5.2: Crear LabelEncoder para mantener consistencia
        self.label_encoder = LabelEncoder()
        y_train_encoded = self.label_encoder.fit_transform(y_train_flat)
        y_test_encoded = self.label_encoder.transform(y_test_flat)

        # ---- 2.5.3: Convertir a one-hot encoding
        self.y_train_onehot = tf.keras.utils.to_categorical(y_train_encoded, num_classes=self.num_classes)
        self.y_test_onehot = tf.keras.utils.to_categorical(y_test_encoded, num_classes=self.num_classes)

        # ---- 2.5.4: Actualizar los datasets de entrenamiento y prueba
        self._update_datasets_with_one_hot()

        self.CT.IT.info_print(f"Targets convertidos a one-hot encoding. Forma y_train: {self.y_train_onehot.shape}, y_test: {self.y_test_onehot.shape}")

    def _update_datasets_with_one_hot(self):
        """
        Actualiza los datasets de TensorFlow con los targets en one-hot encoding
        """
        # ---- 2.5.5: Recrear train_dataset con one-hot encoding
        self.train_dataset = tf.data.Dataset.from_tensor_slices(
            (self.x_train_df.values.astype(np.float32), self.y_train_onehot.astype(np.float32))
        ).batch(self.hiperparams.get('batch_size', 32)).prefetch(tf.data.AUTOTUNE)

        # ---- 2.5.6: Recrear test_dataset con one-hot encoding
        self.test_dataset = tf.data.Dataset.from_tensor_slices(
            (self.x_test_df.values.astype(np.float32), self.y_test_onehot.astype(np.float32))
        ).batch(self.hiperparams.get('batch_size', 32)).prefetch(tf.data.AUTOTUNE)

    # <editor-fold desc="Metodo privado para crear la red    -------------------------------------------------------------------------------------------------">

    def _build_dnn(self, input_shape: tuple) -> tf.keras.Model:
        """
        Construye una DNN avanzada para clasificación multiclase con configuración flexible de capas
        :param input_shape: Forma de los datos de entrada
        :return: Modelo de Keras compilado
        """

        # --------------------------------------------------------------------------------------------
        # -- 1: Obtengo los hipermarámetros y la arquitectura de la red
        # --------------------------------------------------------------------------------------------

        # ---- 1.1: Hiperparámetros básicos (vienen de hiperparams y deben ser numericos)
        units = self.hiperparams.get('units', 64)
        dropout_rate = self.hiperparams.get('dropout', 0.3)
        learning_rate = self.hiperparams.get('learning_rate', 0.001)
        regularization = self.hiperparams.get('regularization', 0.01)

        # ---- 1.2: Hiperparámetros de configuracion (se pasan como parametro individualmente)
        hidden_layers = self.hidden_layers
        activation = self.activation
        geometry = self.geometry_symbol  # '<>', '><', '>', '<'
        include_dropout = self.include_dropout
        include_batch_norm = self.include_batch_norm
        optimizer_type = self.optimizer

        # ---- 1.3: Regularizador (L1 o L2)
        if self.regularization_type == 'l1':
            regularizer = tf.keras.regularizers.L1(regularization)
        else:
            regularizer = tf.keras.regularizers.L2(regularization)

        # --------------------------------------------------------------------------------------------
        # -- 2: Pinto arquitectura y creo el objeto model
        # --------------------------------------------------------------------------------------------

        # ---- 2.1: Pinto la arquitectura de la red
        self.CT.IT.sub_intro_print("NetworkSpecifications")
        self.CT.IT.info_print(f"Arquitectura de red: {geometry}")
        self.CT.IT.info_print(f"Capas: entrada + {hidden_layers} ocultas + salida")
        self.CT.IT.info_print(f"Unidades base: {units}, Dropout: {dropout_rate}, Activación: {activation}")
        self.CT.IT.info_print(f"Número de clases: {self.num_classes}")

        # ---- 2.2: Calculao el punto de inflexión para geometrías simétricas
        inflection_point = hidden_layers // 2

        # ---- 2.3: Defino el modelo vacío
        model = tf.keras.Sequential(name="advanced_dnn_multiclass_classifier")

        # --------------------------------------------------------------------------------------------
        # -- 3: Realizo la configuracion dinamica de la red
        # --------------------------------------------------------------------------------------------

        # ---- 3.1: Agrego capa de activación
        model.add(tf.keras.layers.Dense(
            units=units,
            activation=activation,
            input_shape=input_shape,
            kernel_regularizer=regularizer,
            name="input_dense"
        ))

        # ---- 3.2: Incluyo capa de normalización si así está definido
        if include_batch_norm:
            model.add(tf.keras.layers.BatchNormalization(name="input_bn"))

        # ---- 3.2: Incluyo capa de dropout si así está definido
        if include_dropout:
            model.add(tf.keras.layers.Dropout(dropout_rate, name="input_dropout"))

        # ---- 3.3: Itero para realizar la configuración de las hidden layers
        for i in range(1, hidden_layers + 1):

            # ------ 3.3.1: Defino caracteristicas para evitar el error de reference
            coef: int = 1
            layer_type = "bottleneck"

            # ------ 3.3.2: Hago el matchpara obtener el coeficiente y el tipo de capa
            match geometry:
                case "<>":  # Decrece y luego crece (hourglass)
                    if i <= inflection_point:
                        coef = 2 ** i
                    else:
                        coef = 2 ** ((hidden_layers + 1) - i)
                    layer_type = "bottleneck"

                case "><":  # Crece y luego decrece (diamond)
                    if i <= inflection_point:
                        coef = 2 ** ((hidden_layers + 1) - i)
                    else:
                        coef = 2 ** i
                    layer_type = "expansion"

                case ">":  # Solo decrece (funnel)
                    coef = 2 ** ((hidden_layers + 1) - i)
                    layer_type = "funnel"

                case "<":  # Solo crece (inverse funnel)
                    coef = 2 ** i
                    layer_type = "inverse_funnel"

                case _:  # Constante (fallback)
                    coef = 1
                    layer_type = "constant"

            # ------ 3.3.3: Calculo las unidades para esta capa (asigno un minimo de 8)
            current_units = max(int(units * coef), 8)

            # ------ 3.3.4: Agrego la capa al modelo
            model.add(tf.keras.layers.Dense(
                units=current_units,
                activation=activation,
                kernel_regularizer=regularizer,
                name=f"hidden_{layer_type}_{i}_dense"
            ))

            # ------ 3.3.5: Agrego batch normalization si esta dispuesto así
            if include_batch_norm:
                model.add(tf.keras.layers.BatchNormalization(name=f"hidden_bn_{i}"))

            # ------ 3.3.6: Agrego dropout si esta dispuesto así
            if include_dropout:
                model.add(tf.keras.layers.Dropout(dropout_rate, name=f"hidden_dropout_{i}"))

            # ------ 3.3.7: Pinto la arquitectura de la capa
            self.CT.IT.info_print(f"Capa {i}: {current_units} unidades (coef: {coef})", "cyan")

        # ---- 3.4: Agrego la capa de salida (MODIFICADO PARA MULTICLASE)
        model.add(tf.keras.layers.Dense(
            units=self.num_classes,  # N neuronas para clasificación multiclase
            activation='softmax',  # Softmax para multiclase
            name="output_dense"
        ))

        # --------------------------------------------------------------------------------------------
        # -- 4: Pinto el resumen de la arquitectura, selecciono optimizador y compilo
        # --------------------------------------------------------------------------------------------

        # ---- 4.1: Obtengo los parametros y pinto
        total_params = model.count_params()
        self.CT.IT.info_print(f"Arquitectura finalizada: {geometry} con {hidden_layers} capas ocultas")
        self.CT.IT.info_print(f"Total de parámetros: {total_params:,}")

        # ---- 4.2: Selecciono el optimizador
        if optimizer_type.lower() == 'adam':
            optimizer = tf.keras.optimizers.Adam(
                learning_rate=learning_rate,
                beta_1=0.9,
                beta_2=0.999,
                epsilon=1e-7,
                amsgrad=False
            )
        elif optimizer_type.lower() == 'rmsprop':
            optimizer = tf.keras.optimizers.RMSprop(
                learning_rate=learning_rate,
                rho=0.9,
                momentum=0.0,
                epsilon=1e-7
            )
        else:
            optimizer = tf.keras.optimizers.SGD(
                learning_rate=learning_rate,
                momentum=0.9,
                nesterov=True
            )

        # ---- 4.3: Agrego las métricas (MODIFICADO PARA MULTICLASE)
        metrics = [
            'accuracy',
            tf.keras.metrics.Precision(name='precision'),
            tf.keras.metrics.Recall(name='recall'),
            tf.keras.metrics.CategoricalAccuracy(name='categorical_accuracy'),
        ]

        # ---- 4.4: Compilo el modelo (MODIFICADO PARA MULTICLASE)
        model.compile(
            optimizer=optimizer,
            loss='categorical_crossentropy',  # Categorical crossentropy para multiclase
            metrics=metrics
        )

        # ---- 4.5: Retorno el modelo
        return model

    # </editor-fold>

    # <editor-fold desc="Implementacion de metodos abstractos">

    def get_save_path(self) -> str:
        """
        Metodo que crea la estructura de carpetas para guardar el modelo
        """
        if not os.path.exists(f"{self.save_path}/{self.current_train_version}/{self.model_name}"):
            self.CT.OT.create_folder_if_not_exists(f"{self.save_path}/{self.current_train_version}/{self.model_name}")
        else:
            self.current_train_version: str = self.new_training_version()
            self.CT.OT.create_folder_if_not_exists(f"{self.save_path}/{self.current_train_version}")
            self.CT.OT.create_folder_if_not_exists(f"{self.save_path}/{self.current_train_version}/{self.model_name}")

        self.day_train_versions: List[str] = self.list_versions()
        return f"{self.save_path}/{self.current_train_version}/{self.model_name}"

    def fit_and_get_metrics(self, save_metrics_dict: bool = True) -> Dict[str, Any]:
        """
        Entrena la red neuronal y devuelve las metricas
        :param save_metrics_dict:
        :return:
        """

        # --------------------------------------------------------------------------------------------
        # -- 0: Pinto la entrada
        # --------------------------------------------------------------------------------------------

        start_date: datetime.datetime = datetime.datetime.now()
        self.CT.IT.sub_intro_print(f"{self.class_name}.fit_and_get_metrics: Inicio de entrenamiento a las {start_date}...")

        # --------------------------------------------------------------------------------------------
        # -- 1: Construyo el modelo y entreno
        # --------------------------------------------------------------------------------------------

        try:
            # ---- 1.1: Obtengo el batch sample, el feature data y el inpit_shape del tensor
            sample_batch = next(iter(self.train_dataset.take(1)))
            feature_data = sample_batch[0]
            input_shape = (feature_data.shape[-1],)

            # ---- 1.2: Buildeo el modelo
            self.model = self._build_dnn(input_shape)

            # ---- 1.3: Pinto la arquitectura del mddelo si el verbose está en True
            if self.verbose:
                print(self.model.summary())
                self.CT.IT.info_print(f"Modelo creado con {self.model.count_params():,} parámetros")

            # ---- 1.4: Realizo el entrenamiento (AHORA CON ONE-HOT ENCODING)
            self.model.fit(
                self.train_dataset,
                epochs=self.hiperparams.get('epochs', 100),
                validation_data=self.test_dataset,
                callbacks=self.callbacks,
                verbose=self.verbose,
                shuffle=False
            )

            # --------------------------------------------------------------------------------------------
            # -- 2: Realizo las predicciones y formateo los resultados
            # --------------------------------------------------------------------------------------------

            # ---- 2.1: Realizo las predicciones
            y_pred_train_proba = self.model.predict(self.x_train_df.values, verbose=0)
            y_pred_test_proba = self.model.predict(self.x_test_df.values, verbose=0)

            # ---- 2.2: Obtengo las clases predichas (índice de la máxima probabilidad)
            y_pred_train_classes = np.argmax(y_pred_train_proba, axis=1)
            y_pred_test_classes = np.argmax(y_pred_test_proba, axis=1)

            # ---- 2.3: Convertir one-hot encoding de vuelta a labels originales para evaluación
            y_train_flat = self.y_train.values.ravel() if len(self.y_train.shape) > 1 else self.y_train
            y_test_flat = self.y_test.values.ravel() if len(self.y_test.shape) > 1 else self.y_test

            # --------------------------------------------------------------------------------------------
            # -- 3: Obtengo métricas usando EvaluateMulticlassClassifier
            # --------------------------------------------------------------------------------------------

            # ---- 3.1: Ejecuto la evaluación con EvaluateMulticlassClassifier
            metrics = EvaluateMulticlassClassifier({
                "y_train": y_train_flat,  # Usar labels originales, no one-hot
                "y_eval": None,
                "y_test": y_test_flat,    # Usar labels originales, no one-hot
                "y_pred_train": y_pred_train_classes,
                "y_pred_eval": None,
                "y_pred_test": y_pred_test_classes,
                "target_col_name": self.target_col_name
            }, self.model_name, self.metric, n_classes=self.num_classes).calculate_and_print_metrics()

            # ---- 3.2: Guardo el dict de metricas si asi está especificado
            if save_metrics_dict:
                self.save_metrics_dict_json(metrics, self.model_save_path)

            # ---- 3.3: Pinto la salida
            end_date = datetime.datetime.now()
            self.CT.IT.info_print(f"fit_and_get_metrics terminado a las {end_date}. Duración: {end_date - start_date}", "light_magenta")

            # ---- 3.4: Retorno
            return metrics

        except Exception as e:
            self.CT.IT.info_print(f"Error durante el entrenamiento: {e}")
            raise

    def fit_and_get_model_and_results(self) -> Tuple[Any, np.ndarray, np.ndarray, Dict[str, Any], List]:
        """
        Entrena y devuelve el modelo completo con resultados
        :return: Tupla con (modelo, probabilidades test, probabilidades train, métricas, pesos)
        """
        # --------------------------------------------------------------------------------------------
        # -- 1: Pinto entrada y utilizo el metodo self.fit_and_get_metrics para entrenar el modelo
        # --------------------------------------------------------------------------------------------

        # ---- 1.1: Pinto la entrada para tener referencia de lo que tarda
        start_date: datetime.datetime = datetime.datetime.now()
        self.CT.IT.sub_intro_print(f"{self.class_name}.fit_and_get_model_and_results: Inicio de entrenamiento a las {start_date}...")

        # ---- 1.2: Entreno el modelo y evalúo metricas usando self.fit_and_get_metrics(save_metrics_dict=False)
        metrics = self.fit_and_get_metrics(save_metrics_dict=False)

        # --------------------------------------------------------------------------------------------
        # -- 2: Realizo las predicciones y obtengo las probabilidades de todas las clases
        # --------------------------------------------------------------------------------------------

        # ---- 2.1: Realizo predicciones sobre el conjunto de train y test
        x_train_probs = self.model.predict(self.x_train_df.values, verbose=0)
        x_test_probs = self.model.predict(self.x_test_df.values, verbose=0)

        # ---- 2.2: Para multiclase, mantenemos todas las probabilidades (n, num_classes)
        x_train_probs: np.ndarray = x_train_probs
        x_test_probs: np.ndarray = x_test_probs

        # --------------------------------------------------------------------------------------------
        # -- 3: Obtengo los pesos de la primera capa y preparo los datos para MulticlassClassifierMetricsCalculations
        # --------------------------------------------------------------------------------------------

        # ---- 3.1: Obtengo los pesos
        model_weights, model_weights_dict = self.get_weights(self.model)

        # ---- 3.2: Obtengo las clases predichas
        y_pred_train_classes = np.argmax(x_train_probs, axis=1)
        y_pred_test_classes = np.argmax(x_test_probs, axis=1)

        # ---- 3.3: Asegurar que las etiquetas reales estén en el formato correcto (labels originales)
        y_train_flat = self.y_train.values.ravel() if len(self.y_train.shape) > 1 else self.y_train
        y_test_flat = self.y_test.values.ravel() if len(self.y_test.shape) > 1 else self.y_test

        # ---- 3.4: Crear el DataFrame con las probabilidades para cada clase
        probs_result_dict = {
            f'real_value_{self.target_col_name}': y_test_flat  # Labels originales
        }

        # Agregar probabilidades para cada clase
        for i in range(self.num_classes):
            probs_result_dict[f'proba_class_{i}'] = x_test_probs[:, i]

        # ---- 3.5: Transformo a dataframe
        probs_result_df = pd.DataFrame(probs_result_dict)

        # ---- 3.6: Evalúo las métricas usando MulticlassClassifierMetricsCalculations
        metrics_dict = MulticlassClassifierMetricsCalculations(
            probs_result_df,
            self.model_save_path,
            self.x_test_df,
            self.model,
            model_weights,
            self.target_col_name,
            self.model_name,
            self.bins,
            n_classes=self.num_classes
        ).run()

        # --------------------------------------------------------------------------------------------
        # -- 4: Construyo el diccionario de salida y retorno la tupla
        # --------------------------------------------------------------------------------------------

        # ---- 4.1: Agrego la informacion al master_result_dict
        self.master_result_dict[f"{self.model_name}"] = {
            "columns": list(self.x_train_df.columns),
            "target": self.target_col_name,
            "best_params": self.hiperparams,
            "geometry": self.geometry,
            "geometry_symbol": self.geometry_symbol,
            "hidden_layers": self.hidden_layers,
            "activation": self.activation,
            "regularization_type": self.regularization_type,
            "optimizer": self.optimizer,
            "include_dropout": self.include_dropout,
            "include_batch_norm": self.include_batch_norm,
            "early_stopping_patience": self.early_stopping_patience,
            "metrics": metrics_dict,
            "model_weights": model_weights_dict,
            "model_class_name": self.model.__class__.__name__,
            "model_module_name": self.model.__class__.__module__,
            "model_architecture": json.loads(self.model.to_json()),
            "model_summary": self.get_model_summary(self.model),
            "num_classes": self.num_classes,
            "label_encoder_classes": self.label_encoder.classes_.tolist(),  # Guardar mapeo de clases
            "model_version": f"{self.today_str_daye}_{self.current_train_version}",
        }

        # ---- 4.2: Guardo el modelo en formato .keras
        self.model.save(os.path.join(self.model_save_path, f'{self.class_name}.keras'))

        # ---- 4.3: Almaceno el self.master_dict
        self.save_metrics_dict_json(self.master_result_dict, self.model_save_path, "complete_train_result")

        # ---- 4.4: Pinto salida y retorno la tupla
        end_date = datetime.datetime.now()
        self.CT.IT.info_print(f"fit_and_get_model_and_results terminado a las {end_date}. Duración: {end_date - start_date}", "light_magenta")

        return self.model, x_test_probs, x_train_probs, self.master_result_dict, model_weights

    def get_model_save_path(self) -> str:
        return self.model_save_path

    def execute_shap_analysis(self, sample: bool = True, num_features_to_show: int = 100, num_sample: int = 200, background_sample: int = 100) -> pd.DataFrame:
        """
        Metodo para ejecutar el analisis SHAP de un modelo de Deep Learning multiclase
        """
        return DlShapToolsMulticlass(
            self.x_test_df,
            self.model_name,
            self.model_save_path,
            self.model,
            sample,
            num_features_to_show,
            num_sample,
            background_sample,
            n_classes=self.num_classes
        ).run()

    # </editor-fold>

    # <editor-fold desc="Implementacion de propiedades abstractas">

    @property
    def verbose(self):
        verbose_dict = {'show_all': 1, 'show_basics': 1, 'dont_show': 0}
        return verbose_dict[self._verbose]

    @property
    def metric(self):
        if self._metric not in ['accuracy', 'precision_macro', 'recall_macro', 'f1_macro', 'f1_micro', 'balanced_accuracy', 'roc_auc_ovr']:
            raise ValueError(f'Invalid metric: {self._metric}')
        return self._metric

    @property
    def problem_type(self):
        if self._problem_type not in ['multiclass']:
            raise ValueError(f'Invalid problem_type: {self._problem_type}. Must be multiclass')
        return self._problem_type

    # </editor-fold>


from hx_machine_learning_tools.binary_classifier_ml_models import HxLightGbmBinaryClassifier
from hx_machine_learning_tools.binary_classifier_ml_models import HxXtremeGradientBoostingBinaryClassifier
from hx_machine_learning_tools.binary_classifier_ml_models import HxCatBoostBinaryClassifier
from hx_machine_learning_tools.binary_classifier_ml_models import HxRandomForestBinaryClassifier

from hx_machine_learning_tools.multiclass_classifier_ml_models import HxLightGbmMulticlassClassifier
from hx_machine_learning_tools.multiclass_classifier_ml_models import HxXtremeGradientBoostingMulticlassClassifier
from hx_machine_learning_tools.multiclass_classifier_ml_models import HxCatBoostMulticlassClassifier

from hx_machine_learning_tools.regressor_ml_models import HxLightGbmRegressor
from hx_machine_learning_tools.regressor_ml_models import HxXtremeGradientBoostingRegressor
from hx_machine_learning_tools.regressor_ml_models import HxCatBoostRegressor


from hx_ml_dl_tools import HxLightGbmMulticlassClassifier, HxXtremeGradientBoostingMulticlassClassifier, HxCatBoostMulticlassClassifier, HxDenseNeuralNetworkMulticlassClassifier
from sklearn.datasets import load_iris, load_wine, load_digits
from sklearn.model_selection import train_test_split
from hx_hyperparameters_builder import HxLightgbmHyperparameterBuilder, HxXgboostHyperparameterBuilder, HxCatBoostHyperparameterBuilder, HxDnnHyperparameterBuilder
from typing import Dict, Union
from info_tools import InfoTools
import pandas as pd
import numpy as np


class IrisMulticlassExample:
    def __init__(self):
        """
        Ejemplo de clasificación MULTICLASE de modelos de ML para el dataset Iris de sklearn
        """

        # --------------------------------------------------------------------------------------
        # -- 0: Instancio InfoTools y pinto la entrada
        # --------------------------------------------------------------------------------------

        # ---- 0.1: Instancia de InfoTools
        self.IT: InfoTools = InfoTools()

        # ---- 0.2: Pinto la entrada
        self.IT.header_print(f"IrisMulticlassExample: Ejemplo de clasificación MULTICLASE con sklearn.datasets.iris utilizando hx_ml_dl_tools")

        # --------------------------------------------------------------------------------------
        # -- 1: Descargo el dataset multiclase y obtengo el data dict
        # --------------------------------------------------------------------------------------

        # ---- 1.1: Descargo el dataset Iris (3 clases)
        self.iris_dataset = load_iris()

        # ---- 1.2: Transformo a dataframe (Lo requieren los modelos)
        self.x = pd.DataFrame(self.iris_dataset.data, columns=self.iris_dataset.feature_names)
        self.y = pd.DataFrame(self.iris_dataset.target, columns=["target"])

        # ---- 1.3: Divido en train y test con estratificación para multiclase
        self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(
            self.x, self.y, test_size=0.2, random_state=42, stratify=self.y
        )

        # ---- 1.4: Creo el diccionario en el formato requerido por los modelos
        self.data_dict = {
            "x_train": self.x_train,
            "y_train": self.y_train,
            "x_test": self.x_test,
            "y_test": self.y_test
        }

        # ---- 1.5: Número de clases
        self.n_classes = len(np.unique(self.y))

        # --------------------------------------------------------------------------------------
        # -- 2: Obtengo informacion basica del dataset multiclase
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Información del dataset
        self.IT.sub_intro_print(f"Dataset Iris (Multiclase) cargado:")
        self.IT.info_print(f"Muestras totales: {self.x.shape[0]}")
        self.IT.info_print(f"Características: {self.x.shape[1]}")
        self.IT.info_print(f"Número de clases: {self.n_classes}")
        self.IT.info_print(f"Clases: {self.iris_dataset.target_names.tolist()}")
        self.IT.info_print(f"Train samples: {self.x_train.shape[0]}")
        self.IT.info_print(f"Test samples: {self.x_test.shape[0]}")

        # ---- 2.2: Informacion de distribucion
        self.get_dataset_info()

    def execute_example(self):
        """
        Metodo que ejecuta el entrenamiento multiclase
        :return:
        """
        self._light_gbm_multiclass_classifier()
        self._xg_boost_multiclass_classifier()
        self._cat_boost_multiclass_classifier()
        self._dnn_multiclass_classifier()

    def get_dataset_info(self):
        """
        Metodo adicional para obtener información detallada del dataset multiclase
        """
        self.IT.sub_intro_print("Estadísticas del target (Multiclase):")

        # Distribución por clase
        class_counts = self.y['target'].value_counts().sort_index()
        class_percentages = (class_counts / len(self.y)) * 100

        self.IT.info_print("Distribución por clase:")
        for class_idx, count in class_counts.items():
            class_name = self.iris_dataset.target_names[class_idx]
            percentage = class_percentages[class_idx]
            self.IT.info_print(f"  - {class_name} (Clase {class_idx}): {count} muestras ({percentage:.1f}%)")

        self.IT.info_print(f"Balance: {'Balanceado' if class_percentages.std() < 10 else 'Desbalanceado'}")

        # Estadísticas generales
        self.IT.info_print(f"Media target: {self.y['target'].mean():.2f}")
        self.IT.info_print(f"Std target: {self.y['target'].std():.2f}")

    # <editor-fold desc="Modelo Multiclase   -------------------------------------------------------------------------------------------------------------------------------------">

    def _light_gbm_multiclass_classifier(self):
        """
        Metodo para realizar el entrenamiento del LGBM para clasificación MULTICLASE
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar (adaptados para multiclase)
        # --------------------------------------------------------------------------------------

        # ---- 1.1: Hiperparámetros específicos para multiclase
        lgbm_hyperparams: Dict[str, Union[int, float]] = HxLightgbmHyperparameterBuilder().get_random_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase MULTICLASE que me va a proporcionar un toolkit para ese modelo
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados para multiclase
        lgbm_multiclass: HxLightGbmMulticlassClassifier = HxLightGbmMulticlassClassifier(
            data_dict=self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            hiperparams=lgbm_hyperparams,  # Diccionario con los hiperparametros del LGBM
            metric="f1_macro",  # Métrica específica para multiclase
            problem_type="multiclass",  # Tipo de problema MULTICLASE
            save_path=None,  # Path donde se almacenarán los resultados
            bins=None,  # Bins para estratificación
            verbose="dont_show",  # Nivel de verbosidad
            random_state=42,  # Random state para reproducibilidad
            n_classes=self.n_classes  # Número de clases (CRÍTICO para multiclase)
        )

        # ---- 2.2: Metodo principal que ejecuta el entrenamiento completo
        lgbm_multiclass.fit_and_get_model_and_results()

        # ---- 2.3: Ejecutar análisis SHAP (opcional, puede ser intensivo computacionalmente)

        lgbm_multiclass.execute_shap_analysis(
            sample=True,  # Samplear para hacerlo más rápido
            num_features_to_show=10,  # Mostrar top 10 características
            num_sample=100  # Usar 100 muestras para el análisis
        )

    def _xg_boost_multiclass_classifier(self):
        """
        Metodo para realizar el entrenamiento del LGBM para clasificación MULTICLASE
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar (adaptados para multiclase)
        # --------------------------------------------------------------------------------------

        # ---- 1.1: Hiperparámetros específicos para multiclase
        xgb_hyperparams: Dict[str, Union[int, float]] = HxXgboostHyperparameterBuilder().get_random_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase MULTICLASE que me va a proporcionar un toolkit para ese modelo
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados para multiclase
        xgb_multiclass: HxXtremeGradientBoostingMulticlassClassifier = HxXtremeGradientBoostingMulticlassClassifier(
            data_dict=self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            hiperparams=xgb_hyperparams,  # Diccionario con los hiperparametros del LGBM
            metric="f1_macro",  # Métrica específica para multiclase
            problem_type="multiclass",  # Tipo de problema MULTICLASE
            save_path=None,  # Path donde se almacenarán los resultados
            bins=None,  # Bins para estratificación
            verbose="dont_show",  # Nivel de verbosidad
            random_state=42,  # Random state para reproducibilidad
            n_classes=self.n_classes  # Número de clases (CRÍTICO para multiclase)
        )

        # ---- 2.2: Metodo principal que ejecuta el entrenamiento completo
        xgb_multiclass.fit_and_get_model_and_results()

        # ---- 2.3: Ejecutar análisis SHAP (opcional, puede ser intensivo computacionalmente)

        xgb_multiclass.execute_shap_analysis(
            sample=True,  # Samplear para hacerlo más rápido
            num_features_to_show=10,  # Mostrar top 10 características
            num_sample=100  # Usar 100 muestras para el análisis
        )

    def _cat_boost_multiclass_classifier(self):
        """
        Metodo para realizar el entrenamiento del CAT para clasificación MULTICLASE
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar (adaptados para multiclase)
        # --------------------------------------------------------------------------------------

        # ---- 1.1: Hiperparámetros específicos para multiclase
        cat_hyperparams: Dict[str, Union[int, float]] = HxCatBoostHyperparameterBuilder().get_random_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase MULTICLASE que me va a proporcionar un toolkit para ese modelo
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados para multiclase
        cat_multiclass: HxCatBoostMulticlassClassifier = HxCatBoostMulticlassClassifier(
            data_dict=self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            hiperparams=cat_hyperparams,  # Diccionario con los hiperparametros del LGBM
            metric="f1_macro",  # Métrica específica para multiclase
            problem_type="multiclass",  # Tipo de problema MULTICLASE
            save_path=None,  # Path donde se almacenarán los resultados
            bins=None,  # Bins para estratificación
            verbose="dont_show",  # Nivel de verbosidad
            random_state=42,  # Random state para reproducibilidad
            n_classes=self.n_classes  # Número de clases (CRÍTICO para multiclase)
        )

        # ---- 2.2: Metodo principal que ejecuta el entrenamiento completo
        cat_multiclass.fit_and_get_model_and_results()

        # ---- 2.3: Ejecutar análisis SHAP (opcional, puede ser intensivo computacionalmente)

        """cat_multiclass.execute_shap_analysis(
            sample=True,  # Samplear para hacerlo más rápido
            num_features_to_show=10,  # Mostrar top 10 características
            num_sample=100  # Usar 100 muestras para el análisis
        )"""

    def _dnn_multiclass_classifier(self):
        """
        Metodo para realizar el entrenamiento de la red neuronal densa de clasificacion binaria
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar
        # --------------------------------------------------------------------------------------

        # ---- 1.1: Opcion de pasarselos a mano
        # dnn_hyperparams: Dict[str, Union[int, float]] = {'units': 8, 'dropout': 0.5, 'learning_rate': 0.01, 'regularization': 0.1,  'epochs': 200}

        # ---- 1.2: Opcion de usar la clase HxDnnHyperparameterBuilder

        # ------ 1.2.1: Obtener parámetros aleatorios dentro del rango preestablecido en HxDnnHyperparameterBuilder
        # dnn_hyperparams: Dict[str, Union[int, float]] = HxDnnHyperparameterBuilder().get_random_hyperparams()

        # ------ 1.2.2: Pasarle los hiperparametros al constructor para que valide si hay alguno que se usa poco o no existe (USAMOS ESTA PARA EL EJEMPLO)
        dnn_hyperparams = HxDnnHyperparameterBuilder({'units': 8, 'dropout': 0.5, 'learning_rate': 0.01, 'regularization': 0.1,  'epochs': 200}).validate_and_get_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase que me va a proporcionar un toolkit para ese modelo en concreto
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados (Entrar en la clase para ver detalle)
        dnn: HxDenseNeuralNetworkMulticlassClassifier = HxDenseNeuralNetworkMulticlassClassifier(
            self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            dnn_hyperparams,  # Diccionario de hiperparametros
            'accuracy',  # ['accuracy', 'precision_macro', 'recall_macro', 'f1_macro', 'f1_micro', 'balanced_accuracy', 'roc_auc_ovr']
            'multiclass',  # Literal['binary', 'multiclass']
            1,  # Literal[0, 1, 2, 3] = 1 ({0: "<", 1: ">", 2: "<>", 3: "><"}) Arquitectura de la red
            1,  # Capas ocultas de la red (Toda red tiene 1 capa de entrada + N capas ocultas + 1 capa de salida)
            "relu",  # Literal['relu', 'tanh', 'sigmoid'] = 'relu' (Funcion de activacion)
            "l1",  # Literal['l1', 'l2'] = 'l2'  (Tipo de regularizacion que se aplicará en caso de aplicar regularizacion)
            "adam",  # Literal['sgd', 'adam', 'rmsprop'] = 'adam'  (Optimizador que se va a utilizar)
            True,  # bool (Incluir o no capas de dropout (en caso de no incluirse, el hiperparametro dropout no hace nada))
            True,  # bool (Incluir capa de batch normalization)
            15,  # Si en N epochs no ha mejorado la métrica, aplica un earlyStop callback y devuelve el mejor modelo hasta la fecha
            None,  # Path donde se almacenarán los resultados (por defecto None, ya que se crean internamente las estructuras de carpetas)
            None,  #  dict | None = None (Diccionario de estratificacion, en binary está por defecto entre 0 y 1)
            "dont_show",  # Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show' (Cantidad de informacion adicional que se muestra en consola)
            42,  # Random state de incializacion de pesos para garantizar reproducibilidad
        )

        # ---- 2.2: [Metodo]: Permite entrenar el modelo, pintar las metricas en pantalla y devolver un diccionario de las metricas
        # dnn.fit_and_get_metrics(save_metrics_dict=False)

        # ---- 2.3: [Metodo]: Metodo principal de la clase que ejecuta internamente 'fit_and_get_metrics', almacena el modelo junto con graficos e info adicional
        dnn.fit_and_get_model_and_results()

        # ---- 2.4: [Metodo]: Permite ejecutar el analisis SHAP del modelo
        dnn.execute_shap_analysis(
            sample=True,  # Realizar un sampleo de la data antes de calcular SHAP (!! Es muy intensivo en cómputo si no se samplea !!)
            num_features_to_show=100,  # En los gráficos se van a mostrar las N características mas importantes (con mas peso)
            num_sample=200,  # Cantidad de muestra (filas) con las que se va a realizar el analisis
            background_sample=100,  # Número de instancias para background (explicador deep/grad).
        )

    # </editor-fold>


class WineQualityMulticlassExample:
    def __init__(self):
        """
        Ejemplo de clasificación MULTICLASE con dataset de Vinos (sklearn)
        """
        # --------------------------------------------------------------------------------------
        # -- 0: Instancio InfoTools y pinto la entrada
        # --------------------------------------------------------------------------------------
        self.IT: InfoTools = InfoTools()
        self.IT.header_print(f"WineQualityMulticlassExample: Ejemplo de clasificación MULTICLASE con sklearn.datasets.wine utilizando hx_ml_dl_tools")

        # --------------------------------------------------------------------------------------
        # -- 1: Descargo el dataset multiclase y obtengo el data dict
        # --------------------------------------------------------------------------------------
        self.wine_dataset = load_wine()
        self.x = pd.DataFrame(self.wine_dataset.data, columns=self.wine_dataset.feature_names)
        self.y = pd.DataFrame(self.wine_dataset.target, columns=["target"])

        self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(
            self.x, self.y, test_size=0.2, random_state=42, stratify=self.y
        )

        self.data_dict = {
            "x_train": self.x_train,
            "y_train": self.y_train,
            "x_test": self.x_test,
            "y_test": self.y_test
        }
        self.n_classes = len(np.unique(self.y))

        # --------------------------------------------------------------------------------------
        # -- 2: Obtengo informacion basica del dataset multiclase
        # --------------------------------------------------------------------------------------
        self.IT.sub_intro_print(f"Dataset Wine (Multiclase) cargado:")
        self.IT.info_print(f"Muestras totales: {self.x.shape[0]}")
        self.IT.info_print(f"Características: {self.x.shape[1]}")
        self.IT.info_print(f"Número de clases: {self.n_classes}")
        self.IT.info_print(f"Clases: {self.wine_dataset.target_names.tolist()}")
        self.IT.info_print(f"Train samples: {self.x_train.shape[0]}")
        self.IT.info_print(f"Test samples: {self.x_test.shape[0]}")

        self.get_dataset_info()

    def execute_example(self):
        """
        Metodo que ejecuta el entrenamiento multiclase
        """
        self._light_gbm_multiclass_classifier()
        self._xg_boost_multiclass_classifier()
        self._cat_boost_multiclass_classifier()
        self._dnn_multiclass_classifier()

    def get_dataset_info(self):
        """
        Metodo adicional para obtener información detallada del dataset multiclase
        """
        self.IT.sub_intro_print("Estadísticas del target (Multiclase):")
        class_counts = self.y['target'].value_counts().sort_index()
        class_percentages = (class_counts / len(self.y)) * 100

        self.IT.info_print("Distribución por clase:")
        for class_idx, count in class_counts.items():
            class_name = self.wine_dataset.target_names[class_idx]
            percentage = class_percentages[class_idx]
            self.IT.info_print(f"  - {class_name} (Clase {class_idx}): {count} muestras ({percentage:.1f}%)")

        self.IT.info_print(f"Balance: {'Balanceado' if class_percentages.std() < 10 else 'Desbalanceado'}")
        self.IT.info_print(f"Media target: {self.y['target'].mean():.2f}")
        self.IT.info_print(f"Std target: {self.y['target'].std():.2f}")

    # <editor-fold desc="Modelo Multiclase ------------------------------------------------------------------------------------------------------------------------------------->

    def _light_gbm_multiclass_classifier(self):
        lgbm_hyperparams = HxLightgbmHyperparameterBuilder().get_random_hyperparams()
        lgbm_multiclass = HxLightGbmMulticlassClassifier(
            data_dict=self.data_dict, hiperparams=lgbm_hyperparams, metric="f1_macro",
            problem_type="multiclass", save_path=None, bins=None, verbose="dont_show",
            random_state=42, n_classes=self.n_classes
        )
        lgbm_multiclass.fit_and_get_model_and_results()
        lgbm_multiclass.execute_shap_analysis(sample=True, num_features_to_show=10, num_sample=100)

    def _xg_boost_multiclass_classifier(self):
        xgb_hyperparams = HxXgboostHyperparameterBuilder().get_random_hyperparams()
        xgb_multiclass = HxXtremeGradientBoostingMulticlassClassifier(
            data_dict=self.data_dict, hiperparams=xgb_hyperparams, metric="f1_macro",
            problem_type="multiclass", save_path=None, bins=None, verbose="dont_show",
            random_state=42, n_classes=self.n_classes
        )
        xgb_multiclass.fit_and_get_model_and_results()
        xgb_multiclass.execute_shap_analysis(sample=True, num_features_to_show=10, num_sample=100)

    def _cat_boost_multiclass_classifier(self):
        cat_hyperparams = HxCatBoostHyperparameterBuilder().get_random_hyperparams()
        cat_multiclass = HxCatBoostMulticlassClassifier(
            data_dict=self.data_dict, hiperparams=cat_hyperparams, metric="f1_macro",
            problem_type="multiclass", save_path=None, bins=None, verbose="dont_show",
            random_state=42, n_classes=self.n_classes
        )
        cat_multiclass.fit_and_get_model_and_results()
        # cat_multiclass.execute_shap_analysis(sample=True, num_features_to_show=10, num_sample=100)

    def _dnn_multiclass_classifier(self):
        dnn_hyperparams = HxDnnHyperparameterBuilder(
            {'units': 16, 'dropout': 0.5, 'learning_rate': 0.01, 'regularization': 0.1, 'epochs': 150}
        ).validate_and_get_hyperparams()
        dnn = HxDenseNeuralNetworkMulticlassClassifier(
            self.data_dict, dnn_hyperparams, 'accuracy', 'multiclass', 1, 1,
            "relu", "l1", "adam", True, True, 15, None, None, "dont_show", 42
        )
        dnn.fit_and_get_model_and_results()
        dnn.execute_shap_analysis(sample=True, num_features_to_show=100, num_sample=200, background_sample=100)
    # </editor-fold>


class DigitsMulticlassExample:
    def __init__(self):
        """
        Ejemplo de clasificación MULTICLASE con dataset de Dígitos escritos a mano (10 clases)
        """
        self.IT: InfoTools = InfoTools()
        self.IT.header_print(f"DigitsMulticlassExample: Ejemplo de clasificación MULTICLASE con sklearn.datasets.digits utilizando hx_ml_dl_tools")

        # --------------------------------------------------------------------------------------
        # -- 1: Dataset multiclase
        # --------------------------------------------------------------------------------------
        self.digits_dataset = load_digits()
        self.x = pd.DataFrame(self.digits_dataset.data, columns=[f"pixel_{i}" for i in range(self.digits_dataset.data.shape[1])])
        self.y = pd.DataFrame(self.digits_dataset.target, columns=["target"])

        self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(
            self.x, self.y, test_size=0.2, random_state=42, stratify=self.y
        )

        self.data_dict = {
            "x_train": self.x_train,
            "y_train": self.y_train,
            "x_test": self.x_test,
            "y_test": self.y_test
        }
        self.n_classes = len(np.unique(self.y))

        # --------------------------------------------------------------------------------------
        # -- 2: Info dataset
        # --------------------------------------------------------------------------------------
        self.IT.sub_intro_print(f"Dataset Digits (Multiclase) cargado:")
        self.IT.info_print(f"Muestras totales: {self.x.shape[0]}")
        self.IT.info_print(f"Características: {self.x.shape[1]}")
        self.IT.info_print(f"Número de clases: {self.n_classes}")
        self.IT.info_print(f"Train samples: {self.x_train.shape[0]}")
        self.IT.info_print(f"Test samples: {self.x_test.shape[0]}")

        self.get_dataset_info()

    def execute_example(self):
        self._light_gbm_multiclass_classifier()
        self._xg_boost_multiclass_classifier()
        self._cat_boost_multiclass_classifier()
        self._dnn_multiclass_classifier()

    def get_dataset_info(self):
        self.IT.sub_intro_print("Distribución por clase (dígitos):")
        class_counts = self.y['target'].value_counts().sort_index()
        for digit, count in class_counts.items():
            self.IT.info_print(f"  - Dígito {digit}: {count} muestras ({(count/len(self.y))*100:.1f}%)")

    # <editor-fold desc="Modelos Multiclase ------------------------------------------------------------------------------------------------------------------------------------->

    def _light_gbm_multiclass_classifier(self):
        lgbm_hyperparams = HxLightgbmHyperparameterBuilder().get_random_hyperparams()
        lgbm_multiclass = HxLightGbmMulticlassClassifier(
            data_dict=self.data_dict, hiperparams=lgbm_hyperparams, metric="f1_macro",
            problem_type="multiclass", save_path=None, bins=None, verbose="dont_show",
            random_state=42, n_classes=self.n_classes
        )
        lgbm_multiclass.fit_and_get_model_and_results()
        lgbm_multiclass.execute_shap_analysis(sample=True, num_features_to_show=10, num_sample=100)

    def _xg_boost_multiclass_classifier(self):
        xgb_hyperparams = HxXgboostHyperparameterBuilder().get_random_hyperparams()
        xgb_multiclass = HxXtremeGradientBoostingMulticlassClassifier(
            data_dict=self.data_dict, hiperparams=xgb_hyperparams, metric="f1_macro",
            problem_type="multiclass", save_path=None, bins=None, verbose="dont_show",
            random_state=42, n_classes=self.n_classes
        )
        xgb_multiclass.fit_and_get_model_and_results()
        xgb_multiclass.execute_shap_analysis(sample=True, num_features_to_show=10, num_sample=100)

    def _cat_boost_multiclass_classifier(self):
        cat_hyperparams = HxCatBoostHyperparameterBuilder().get_random_hyperparams()
        cat_multiclass = HxCatBoostMulticlassClassifier(
            data_dict=self.data_dict, hiperparams=cat_hyperparams, metric="f1_macro",
            problem_type="multiclass", save_path=None, bins=None, verbose="dont_show",
            random_state=42, n_classes=self.n_classes
        )
        cat_multiclass.fit_and_get_model_and_results()
        # cat_multiclass.execute_shap_analysis(sample=True, num_features_to_show=10, num_sample=100)

    def _dnn_multiclass_classifier(self):
        dnn_hyperparams = HxDnnHyperparameterBuilder(
            {'units': 32, 'dropout': 0.5, 'learning_rate': 0.01, 'regularization': 0.1, 'epochs': 100}
        ).validate_and_get_hyperparams()
        dnn = HxDenseNeuralNetworkMulticlassClassifier(
            self.data_dict, dnn_hyperparams, 'accuracy', 'multiclass', 1, 2,
            "relu", "l2", "adam", True, True, 15, None, None, "dont_show", 42
        )
        dnn.fit_and_get_model_and_results()
        dnn.execute_shap_analysis(sample=True, num_features_to_show=100, num_sample=200, background_sample=100)
    # </editor-fold>



from imblearn.under_sampling import TomekLinks, RandomUnderSampler, CondensedNearestNeighbour, NeighbourhoodCleaningRule
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OrdinalEncoder, StandardScaler, RobustScaler, MinMaxScaler, MaxAbsScaler, OneHotEncoder
from category_encoders import TargetEncoder, CatBoostEncoder, JamesSteinEncoder, LeaveOneOutEncoder, SumEncoder
from imblearn.over_sampling import SMOTE
from singleton_tools import SingletonMeta
from typing import Tuple, Optional
import pandas as pd


class SkLearnTools(metaclass=SingletonMeta):
    def __init__(self):
        pass

    # <editor-fold desc="Split    ------------------------------------------------------------------------------------------------------------------------------------------------">

    @staticmethod
    def split_train_test_forecasting(df: pd.DataFrame, date_colname: str, limit_date_str: Optional[str] = None, test_size: float = 0.2) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Divide un DataFrame en conjuntos de entrenamiento y prueba para forecasting.

        Existen dos modos de partición:
        1. Si se proporciona `limit_date_str`, se usará como fecha de corte.
           - Todas las filas con fecha <= límite irán al conjunto de entrenamiento.
           - Todas las filas con fecha > límite irán al conjunto de prueba.
        2. Si no se especifica `limit_date_str`, se dividirá el dataset
           automáticamente según el porcentaje definido en `test_size`.

        :param df: (pd.DataFrame): DataFrame de entrada que contiene los datos.
        :param date_colname: (str): Nombre de la columna que contiene las fechas.
        :param limit_date_str: (Optional[str]): Fecha límite en formato 'YYYY-MM-DD'. Si es None, se usará `test_size`.
        :param test_size: (float, optional): Proporción de datos para el conjunto de prueba (valor entre 0 y 1). Default = 0.2.
        :return: Tuple[pd.DataFrame, pd.DataFrame]
        """


        # Caso 1: Partición por fecha límite explícita
        if limit_date_str is not None:
            # Convertimos la fecha límite a tipo datetime
            limit_date = pd.to_datetime(limit_date_str, format="%Y-%m-%d")

            # Entrenamiento: filas hasta la fecha límite (incluida)
            df_train = df[df[date_colname] <= limit_date].copy()

            # Prueba: filas posteriores a la fecha límite
            df_test = df[df[date_colname] > limit_date].copy()

            return df_train, df_test

        # Caso 2: Partición automática por proporción
        # Ordenamos por columna de fecha para mantener la secuencia temporal
        df_sorted = df.sort_values(by=date_colname)

        # Calculamos tamaño del set de entrenamiento
        train_size = int((1 - test_size) * len(df_sorted))

        # División en train y test
        df_train = df_sorted.iloc[:train_size].copy()
        df_test = df_sorted.iloc[train_size:].copy()

        return df_train, df_test

    @staticmethod
    def split_train_test(df: pd.DataFrame, test_size: float = 0.2, shuffle: bool = False, random_state: Optional[int] = 42, stratify_col: Optional[str] = None) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Divide un DataFrame en conjuntos de entrenamiento y prueba.

        Se puede estratificar por una columna específica para mantener
        la distribución de clases en train y test.

        :param df: DataFrame de entrada.
        :param test_size: Proporción de datos para el conjunto de prueba (entre 0 y 1).
        :param shuffle: Indica si se deben mezclar las filas antes de dividir.
        :param random_state: Semilla aleatoria para reproducibilidad.
        :param stratify_col: Nombre de la columna a usar para estratificación. Si es None, no se aplica estratificación.
        :return: Una tupla (df_train, df_test).
        """
        # -------------------------------------------------------------------------------------------------
        # -- 1: Evaluamos si existe columna de estratificacion
        # -------------------------------------------------------------------------------------------------

        # ---- 1.1: Si no existe, retornamos el train test_split con normalidad
        if stratify_col is None:
            return train_test_split(df, test_size=test_size, shuffle=shuffle, random_state=random_state)

        # ---- 1.2: Si existe, la agregamos al parametro stratify y retornamos
        return train_test_split(df, test_size=test_size, shuffle=shuffle, random_state=random_state, stratify=df[stratify_col])

    # </editor-fold>

    # <editor-fold desc="Forecasting    ------------------------------------------------------------------------------------------------------------------------------------------">

    # </editor-fold>

    # <editor-fold desc="Oversamplers    -----------------------------------------------------------------------------------------------------------------------------------------">
    @staticmethod
    def smote_oversampler(x_train: pd.DataFrame, y_train: pd.Series, sampling_prop: Optional[float] = 0.8) -> Tuple[pd.DataFrame, pd.Series]:
        """
        Aplica la técnica SMOTE (Synthetic Minority Over-sampling Technique) para balancear
        las clases en un conjunto de entrenamiento. Utiliza k-NN para generar ejemplos sintéticos
        de la clase minoritaria en lugar de eliminar ejemplos de la clase mayoritaria.

        :param x_train: Conjunto de características de entrenamiento.
        :param y_train: Conjunto de etiquetas de entrenamiento.
        :param sampling_prop: Proporción de muestreo para la clase minoritaria. Si es None, se usa el valor por defecto de SMOTE.
        :return: Una tupla (X_smote, y_smote) con los datos balanceados.
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Configuración del oversampler SMOTE
        # -------------------------------------------------------------------------------------------------
        smote: SMOTE = SMOTE(sampling_strategy=sampling_prop)

        # -------------------------------------------------------------------------------------------------
        # -- 2: Aplicación del metodo fit_resample para generar nuevos ejemplos
        # -------------------------------------------------------------------------------------------------
        x_smote, y_smote = smote.fit_resample(x_train, y_train)

        # -------------------------------------------------------------------------------------------------
        # -- 3: Retorno de los datos balanceados
        # -------------------------------------------------------------------------------------------------
        return x_smote, y_smote

    # </editor-fold>

    # <editor-fold desc="Undersamplers    ----------------------------------------------------------------------------------------------------------------------------------------">
    @staticmethod
    def neighbourhood_cleaning_rule_undersampler(x_train: pd.DataFrame, y_train: pd.Series, sampling_prop: Optional[float] = 0.6) -> Tuple[pd.DataFrame, pd.Series]:
        """
        Aplica Neighborhood Cleaning Rule (NCR) para eliminar ejemplos de la clase mayoritaria
        mal clasificados o rodeados por ejemplos de la clase minoritaria. Posteriormente,
        puede aplicar Random Under-Sampling para ajustar la proporción.

        :param x_train: Conjunto de características de entrenamiento.
        :param y_train: Conjunto de etiquetas de entrenamiento.
        :param sampling_prop: Proporción de muestreo para la clase mayoritaria. Si es None, se aplica únicamente NCR.
        :return: Una tupla (x_resampled, y_resampled) con los datos balanceados.
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Aplicación de Neighborhood Cleaning Rule (NCR)
        # -------------------------------------------------------------------------------------------------
        ncr: NeighbourhoodCleaningRule = NeighbourhoodCleaningRule()
        x_ncr, y_ncr = ncr.fit_resample(x_train, y_train)

        # -------------------------------------------------------------------------------------------------
        # -- 2: Si no se especifica sampling_prop, retornamos directamente NCR
        # -------------------------------------------------------------------------------------------------
        if sampling_prop is None:
            return x_ncr, y_ncr

        # -------------------------------------------------------------------------------------------------
        # -- 3: Aplicamos Random Under-Sampling posterior a NCR
        # -------------------------------------------------------------------------------------------------
        rus: RandomUnderSampler = RandomUnderSampler(sampling_strategy=sampling_prop)
        x_resampled, y_resampled = rus.fit_resample(x_ncr, y_ncr)

        return x_resampled, y_resampled

    @staticmethod
    def condensed_nearest_neighbour_undersampler(x_train: pd.DataFrame, y_train: pd.Series, sampling_prop: Optional[float] = 0.5) -> Tuple[pd.DataFrame, pd.Series]:
        """
        Aplica Condensed Nearest Neighbour (CNN) para seleccionar un subconjunto mínimo
        de ejemplos necesarios para clasificar correctamente la clase minoritaria.
        Posteriormente, puede aplicar Random Under-Sampling para ajustar la proporción.

        :param x_train: Conjunto de características de entrenamiento.
        :param y_train: Conjunto de etiquetas de entrenamiento.
        :param sampling_prop: Proporción de muestreo para la clase mayoritaria. Si es None, se aplica únicamente CNN.
        :return: Una tupla (x_resampled, y_resampled) con los datos balanceados.
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Aplicación de Condensed Nearest Neighbour (CNN)
        # -------------------------------------------------------------------------------------------------
        cnn: CondensedNearestNeighbour = CondensedNearestNeighbour()
        x_cnn, y_cnn = cnn.fit_resample(x_train, y_train)

        # -------------------------------------------------------------------------------------------------
        # -- 2: Si no se especifica sampling_prop, retornamos directamente CNN
        # -------------------------------------------------------------------------------------------------
        if sampling_prop is None:
            return x_cnn, y_cnn

        # -------------------------------------------------------------------------------------------------
        # -- 3: Aplicamos Random Under-Sampling posterior a CNN
        # -------------------------------------------------------------------------------------------------
        rus: RandomUnderSampler = RandomUnderSampler(sampling_strategy=sampling_prop)
        x_resampled, y_resampled = rus.fit_resample(x_cnn, y_cnn)

        return x_resampled, y_resampled

    @staticmethod
    def tomek_links_undersampling(x_train: pd.DataFrame, y_train: pd.Series, sampling_prop: Optional[float] = 0.5) -> Tuple[pd.DataFrame, pd.Series]:
        """
        Aplica Tomek Links, identificando pares de ejemplos de clases opuestas
        que son vecinos más cercanos entre sí. Los ejemplos de la clase mayoritaria
        en cada par se eliminan. Posteriormente, puede aplicar Random Under-Sampling.

        :param x_train: Conjunto de características de entrenamiento.
        :param y_train: Conjunto de etiquetas de entrenamiento.
        :param sampling_prop: Proporción de muestreo para la clase mayoritaria. Si es None, se aplica únicamente Tomek Links.
        :return: Una tupla (x_resampled, y_resampled) con los datos balanceados.
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Aplicación de Tomek Links
        # -------------------------------------------------------------------------------------------------
        tl: TomekLinks = TomekLinks()
        x_tl, y_tl = tl.fit_resample(x_train, y_train)

        # -------------------------------------------------------------------------------------------------
        # -- 2: Si no se especifica sampling_prop, retornamos directamente TL
        # -------------------------------------------------------------------------------------------------
        if sampling_prop is None:
            return x_tl, y_tl

        # -------------------------------------------------------------------------------------------------
        # -- 3: Aplicamos Random Under-Sampling posterior a TL
        # -------------------------------------------------------------------------------------------------
        rus: RandomUnderSampler = RandomUnderSampler(sampling_strategy=sampling_prop)
        x_resampled, y_resampled = rus.fit_resample(x_tl, y_tl)

        return x_resampled, y_resampled

    # </editor-fold>

    # <editor-fold desc="Encoders para aplicar al dataframe completo (sin separar en train y test)    ----------------------------------------------------------------------------">

    @staticmethod
    def ordinal_encoder(df_column: pd.Series, jerarquized_list: list[str]) -> pd.DataFrame:
        """
        Aplica codificación ordinal a una columna categórica jerarquizada.
        Asigna valores numéricos consecutivos a categorías que tienen un orden jerárquico.

        Ejemplo:
            primaria -> 0
            secundaria -> 1
            universidad -> 2

        :param df_column: Columna del DataFrame a codificar.
        :param jerarquized_list: Lista ORDENADA DE MENOR A MAYOR de las categorías.
        :return: Columna transformada como DataFrame con valores ordinales.
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Inicialización del codificador ordinal con la jerarquía definida
        # -------------------------------------------------------------------------------------------------
        ordinal_encoder: OrdinalEncoder = OrdinalEncoder(categories=[jerarquized_list])

        # -------------------------------------------------------------------------------------------------
        # -- 2: Transformación de la columna en un array 2D y aplicación del encoder
        # -------------------------------------------------------------------------------------------------
        encoded_column = ordinal_encoder.fit_transform(df_column.to_numpy().reshape(-1, 1))

        # -------------------------------------------------------------------------------------------------
        # -- 3: Retorno como DataFrame
        # -------------------------------------------------------------------------------------------------
        return pd.DataFrame(encoded_column, columns=[df_column.name])

    @staticmethod
    def one_hot_encoder(df: pd.DataFrame, cols_to_encode: list[str]) -> Tuple[pd.DataFrame, list[str]]:
        """
        Aplica One-Hot Encoding a las columnas seleccionadas de un DataFrame.
        Genera nuevas columnas binarias para cada categoría encontrada.

        :param df: DataFrame de entrada.
        :param cols_to_encode: Lista de columnas a codificar.
        :return: Una tupla (new_df, feature_names).
                 - new_df: DataFrame con las columnas codificadas.
                 - feature_names: Lista de los nombres de las columnas creadas.
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Inicialización del codificador OneHot con soporte para categorías desconocidas
        # -------------------------------------------------------------------------------------------------
        encoder = OneHotEncoder(handle_unknown='ignore', sparse_output=False)

        # -------------------------------------------------------------------------------------------------
        # -- 2: Ajuste del encoder a las categorías presentes en las columnas seleccionadas
        # -------------------------------------------------------------------------------------------------
        encoder.fit(df[cols_to_encode])

        # -------------------------------------------------------------------------------------------------
        # -- 3: Transformación de las columnas categóricas
        # -------------------------------------------------------------------------------------------------
        encoded_cols = encoder.transform(df[cols_to_encode])

        # -------------------------------------------------------------------------------------------------
        # -- 4: Obtención de los nombres de las nuevas columnas
        # -------------------------------------------------------------------------------------------------
        feature_names = encoder.get_feature_names_out(cols_to_encode).tolist()

        # -------------------------------------------------------------------------------------------------
        # -- 5: Creación de un DataFrame con las columnas codificadas
        # -------------------------------------------------------------------------------------------------
        encoded_df = pd.DataFrame(encoded_cols, columns=feature_names, index=df.index)

        # -------------------------------------------------------------------------------------------------
        # -- 6: Concatenación del DataFrame original (sin las columnas originales) con las codificadas
        # -------------------------------------------------------------------------------------------------
        new_df = pd.concat([df.drop(cols_to_encode, axis=1), encoded_df], axis=1)

        return new_df, feature_names

    @staticmethod
    def one_hot_encoder_pd(df: pd.DataFrame, cols_to_encode: list[str]) -> pd.DataFrame:
        """
        Aplica One-Hot Encoding usando pandas.get_dummies a las columnas seleccionadas.
        Genera nuevas columnas binarias para cada categoría encontrada.

        :param df: DataFrame de entrada.
        :param cols_to_encode: Lista de columnas a codificar.
        :return: DataFrame con las columnas codificadas.
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Aplicación de get_dummies a las columnas seleccionadas
        # -------------------------------------------------------------------------------------------------
        encoded_df = pd.get_dummies(df, columns=cols_to_encode)

        return encoded_df

    # </editor-fold>

    # <editor-fold desc="Encoders para aplicar al conjunto de datos separado en train y test    ----------------------------------------------------------------------------------">

    @staticmethod
    def catboost_encoder(train_df: pd.DataFrame, test_df: pd.DataFrame, target: pd.Series, cols_to_encode: list[str]) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Codifica variables categóricas utilizando CatBoostEncoder.
        Similar al Target Encoding pero más robusto frente a overfitting y outliers.

        ➤ Usar cuando: buscas un encoding potente y robusto para modelos de árbol
          y tienes suficiente cantidad de datos por categoría.

        :param train_df: DataFrame de entrenamiento (sin columna objetivo).
        :param test_df: DataFrame de prueba (sin columna objetivo).
        :param target: Serie objetivo de entrenamiento.
        :param cols_to_encode: Lista de columnas a codificar.
        :return: Tupla (train_df_transformado, test_df_transformado).
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Copias y validación básica de columnas
        # -------------------------------------------------------------------------------------------------
        train_df: pd.DataFrame = train_df.copy()
        test_df: pd.DataFrame = test_df.copy()
        present_cols = [c for c in cols_to_encode if c in train_df.columns and c in test_df.columns]

        # ---- 1.1: Si no hay columnas válidas, devolvemos copias sin tocar
        if not present_cols:
            return train_df, test_df

        # -------------------------------------------------------------------------------------------------
        # -- 2: Aseguramos tipo category en las columnas a codificar
        # -------------------------------------------------------------------------------------------------
        for col in present_cols:
            train_df[col] = train_df[col].astype("category")
            test_df[col] = test_df[col].astype("category")

        # -------------------------------------------------------------------------------------------------
        # -- 3: Fit y transform con CatBoostEncoder
        # -------------------------------------------------------------------------------------------------
        encoder: CatBoostEncoder = CatBoostEncoder()
        encoder.fit(train_df[present_cols], target)

        train_df[present_cols] = encoder.transform(train_df[present_cols])
        test_df[present_cols] = encoder.transform(test_df[present_cols])

        # -------------------------------------------------------------------------------------------------
        # -- 4: Retorno
        # -------------------------------------------------------------------------------------------------
        return train_df, test_df

    @staticmethod
    def james_stein_encoder(train_df: pd.DataFrame, test_df: pd.DataFrame, target: pd.Series, cols_to_encode: list[str]) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Codifica variables categóricas usando James-Stein encoder.
        Reduce varianza combinando frecuencia y media del target.

        ➤ Usar cuando: quieres regularizar el encoding y mejorar generalización,
          especialmente en categorías con tamaños variados.

        :param train_df: DataFrame de entrenamiento (sin columna objetivo).
        :param test_df: DataFrame de prueba (sin columna objetivo).
        :param target: Serie objetivo de entrenamiento.
        :param cols_to_encode: Lista de columnas a codificar.
        :return: Tupla (train_df_transformado, test_df_transformado).
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Copias y validación básica de columnas
        # -------------------------------------------------------------------------------------------------
        train_df: pd.DataFrame = train_df.copy()
        test_df: pd.DataFrame = test_df.copy()
        present_cols = [c for c in cols_to_encode if c in train_df.columns and c in test_df.columns]
        if not present_cols:
            return train_df, test_df

        # -------------------------------------------------------------------------------------------------
        # -- 2: Aseguramos tipo category en las columnas a codificar
        # -------------------------------------------------------------------------------------------------
        for col in present_cols:
            train_df[col] = train_df[col].astype("category")
            test_df[col] = test_df[col].astype("category")

        # -------------------------------------------------------------------------------------------------
        # -- 3: Fit_transform en train y transform en test con JamesSteinEncoder
        # -------------------------------------------------------------------------------------------------
        encoder: JamesSteinEncoder = JamesSteinEncoder()
        train_encoded = encoder.fit_transform(train_df[present_cols], target)
        test_encoded = encoder.transform(test_df[present_cols])

        # -------------------------------------------------------------------------------------------------
        # -- 4: Reconstrucción de dataframes (sustituimos columnas originales por codificadas)
        # -------------------------------------------------------------------------------------------------
        train_df = pd.concat([train_df.drop(present_cols, axis=1), train_encoded], axis=1)
        test_df = pd.concat([test_df.drop(present_cols, axis=1), test_encoded], axis=1)

        return train_df, test_df

    @staticmethod
    def leave_one_out_encoder(train_df: pd.DataFrame, test_df: pd.DataFrame, target: pd.Series, cols_to_encode: list[str]) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Codifica variables categóricas usando Leave-One-Out encoding.
        Calcula la media del target excluyendo la observación actual para evitar sesgo.

        ➤ Usar cuando: quieres un encoding que reduzca el sesgo del target respecto a la propia fila,
          pero ten cuidado con categorías con muy pocas observaciones (riesgo overfitting).

        :param train_df: DataFrame de entrenamiento (sin columna objetivo).
        :param test_df: DataFrame de prueba (sin columna objetivo).
        :param target: Serie objetivo de entrenamiento.
        :param cols_to_encode: Lista de columnas a codificar.
        :return: Tupla (train_df_transformado, test_df_transformado).
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Copias y validación básica de columnas
        # -------------------------------------------------------------------------------------------------
        train_df: pd.DataFrame = train_df.copy()
        test_df: pd.DataFrame = test_df.copy()
        present_cols = [c for c in cols_to_encode if c in train_df.columns and c in test_df.columns]
        if not present_cols:
            return train_df, test_df

        # -------------------------------------------------------------------------------------------------
        # -- 2: Aseguramos tipo category en las columnas a codificar
        # -------------------------------------------------------------------------------------------------
        for col in present_cols:
            train_df[col] = train_df[col].astype("category")
            test_df[col] = test_df[col].astype("category")

        # -------------------------------------------------------------------------------------------------
        # -- 3: Fit_transform en train y transform en test con LeaveOneOutEncoder
        # -------------------------------------------------------------------------------------------------
        encoder: LeaveOneOutEncoder = LeaveOneOutEncoder()
        train_encoded = encoder.fit_transform(train_df[present_cols], target)
        test_encoded = encoder.transform(test_df[present_cols])

        # -------------------------------------------------------------------------------------------------
        # -- 4: Reconstrucción de dataframes
        # -------------------------------------------------------------------------------------------------
        train_df = pd.concat([train_df.drop(present_cols, axis=1), train_encoded], axis=1)
        test_df = pd.concat([test_df.drop(present_cols, axis=1), test_encoded], axis=1)

        return train_df, test_df

    @staticmethod
    def sum_encoder(train_df: pd.DataFrame, test_df: pd.DataFrame, cols_to_encode: list[str]) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Codifica variables categóricas utilizando SumEncoder.
        Genera representaciones continuas basadas en suma/agregación de dummy-encodings.

        ➤ Usar cuando: buscas una representación numérica simple y rápida,
          útil para modelos lineales o cuando la granularidad no es crítica.

        :param train_df: DataFrame de entrenamiento (sin columna objetivo).
        :param test_df: DataFrame de prueba (sin columna objetivo).
        :param cols_to_encode: Lista de columnas a codificar.
        :return: Tupla (train_df_transformado, test_df_transformado).
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Copias y validación básica de columnas
        # -------------------------------------------------------------------------------------------------
        train_df: pd.DataFrame = train_df.copy()
        test_df: pd.DataFrame = test_df.copy()
        present_cols = [c for c in cols_to_encode if c in train_df.columns and c in test_df.columns]
        if not present_cols:
            return train_df, test_df

        # -------------------------------------------------------------------------------------------------
        # -- 2: Aplicación de SumEncoder sobre el DataFrame completo (según implementación esperada)
        # -------------------------------------------------------------------------------------------------
        encoder: SumEncoder = SumEncoder(cols=present_cols)
        train_encoded = encoder.fit_transform(train_df)
        test_encoded = encoder.transform(test_df)

        # -------------------------------------------------------------------------------------------------
        # -- 3: Retorno (se asume que encoder devuelve DataFrames completos)
        # -------------------------------------------------------------------------------------------------
        return train_encoded, test_encoded

    @staticmethod
    def target_encoder(train_df: pd.DataFrame, test_df: pd.DataFrame, target: pd.Series, cols_to_encode: list[str]) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Codifica variables categóricas utilizando TargetEncoder (media del target por categoría).

        ➤ Usar cuando: necesitas un encoding simple y rápido; ten en cuenta riesgo de overfitting.

        :param train_df: DataFrame de entrenamiento (sin columna objetivo).
        :param test_df: DataFrame de prueba (sin columna objetivo).
        :param target: Serie objetivo de entrenamiento.
        :param cols_to_encode: Lista de columnas a codificar.
        :return: Tupla (train_df_transformado, test_df_transformado).
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Copias y validación básica de columnas
        # -------------------------------------------------------------------------------------------------
        train_df: pd.DataFrame = train_df.copy()
        test_df: pd.DataFrame = test_df.copy()
        present_cols = [c for c in cols_to_encode if c in train_df.columns and c in test_df.columns]
        if not present_cols:
            return train_df, test_df

        # -------------------------------------------------------------------------------------------------
        # -- 2: Aseguramos tipo category en las columnas a codificar
        # -------------------------------------------------------------------------------------------------
        for col in present_cols:
            train_df[col] = train_df[col].astype("category")
            test_df[col] = test_df[col].astype("category")

        # -------------------------------------------------------------------------------------------------
        # -- 3: Fit y transform con TargetEncoder
        # -------------------------------------------------------------------------------------------------
        encoder: TargetEncoder = TargetEncoder()
        encoder.fit(train_df[present_cols], target)

        train_df[present_cols] = encoder.transform(train_df[present_cols])
        test_df[present_cols] = encoder.transform(test_df[present_cols])

        # -------------------------------------------------------------------------------------------------
        # -- 4: Retorno
        # -------------------------------------------------------------------------------------------------
        return train_df, test_df

    # </editor-fold>

    # <editor-fold desc="Scalers    ----------------------------------------------------------------------------------------------------------------------------------------------">

    @staticmethod
    def min_max_scaler(train_df: pd.DataFrame, test_df: pd.DataFrame, cols_to_scale: list[str], excluded_cols: list[str]) -> tuple[pd.DataFrame, pd.DataFrame, MinMaxScaler]:
        """
        Escala los datos entre 0 y 1 utilizando MinMaxScaler.

        ➤ Usar cuando: quieres normalizar datos a un rango fijo, útil para algoritmos sensibles a magnitudes (p. ej. redes neuronales).

        :param train_df: DataFrame de entrenamiento.
        :param test_df: DataFrame de prueba.
        :param cols_to_scale: Columnas a escalar.
        :param excluded_cols: Columnas a mantener sin escalar.
        :return: Tupla (train_scaled_df, test_scaled_df, scaler).
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Copias y separación de columnas excluidas
        # -------------------------------------------------------------------------------------------------
        train_df: pd.DataFrame = train_df.copy()
        test_df: pd.DataFrame = test_df.copy()
        train_excluded = train_df[excluded_cols]
        test_excluded = test_df[excluded_cols]

        # -------------------------------------------------------------------------------------------------
        # -- 2: Aplicación del MinMaxScaler
        # -------------------------------------------------------------------------------------------------
        scaler: MinMaxScaler = MinMaxScaler()
        train_scaled = scaler.fit_transform(train_df[cols_to_scale])
        test_scaled = scaler.transform(test_df[cols_to_scale])

        train_scaled_df = pd.DataFrame(train_scaled, columns=cols_to_scale, index=train_df.index)
        test_scaled_df = pd.DataFrame(test_scaled, columns=cols_to_scale, index=test_df.index)

        # -------------------------------------------------------------------------------------------------
        # -- 3: Concatenar columnas excluidas
        # -------------------------------------------------------------------------------------------------
        train_scaled_df = pd.concat([train_scaled_df, train_excluded], axis=1)
        test_scaled_df = pd.concat([test_scaled_df, test_excluded], axis=1)

        return train_scaled_df, test_scaled_df, scaler

    @staticmethod
    def max_abs_scaler(train_df: pd.DataFrame, test_df: pd.DataFrame, cols_to_scale: list[str], excluded_cols: list[str]) -> tuple[pd.DataFrame, pd.DataFrame, MaxAbsScaler]:
        """
        Escala los datos de forma que el valor absoluto máximo de cada columna sea 1.

        ➤ Usar cuando: los datos contienen valores positivos y negativos y quieres preservar la dispersión relativa.

        :param train_df: DataFrame de entrenamiento.
        :param test_df: DataFrame de prueba.
        :param cols_to_scale: Columnas a escalar.
        :param excluded_cols: Columnas a mantener sin escalar.
        :return: Tupla (train_scaled_df, test_scaled_df, scaler).
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Copias y separación de columnas excluidas
        # -------------------------------------------------------------------------------------------------
        train_df: pd.DataFrame = train_df.copy()
        test_df: pd.DataFrame = test_df.copy()
        train_excluded = train_df[excluded_cols]
        test_excluded = test_df[excluded_cols]

        # -------------------------------------------------------------------------------------------------
        # -- 2: Aplicación del MaxAbsScaler
        # -------------------------------------------------------------------------------------------------
        scaler: MaxAbsScaler = MaxAbsScaler()
        train_scaled = scaler.fit_transform(train_df[cols_to_scale])
        test_scaled = scaler.transform(test_df[cols_to_scale])

        train_scaled_df = pd.DataFrame(train_scaled, columns=cols_to_scale, index=train_df.index)
        test_scaled_df = pd.DataFrame(test_scaled, columns=cols_to_scale, index=test_df.index)

        # -------------------------------------------------------------------------------------------------
        # -- 3: Concatenar columnas excluidas
        # -------------------------------------------------------------------------------------------------
        train_scaled_df = pd.concat([train_scaled_df, train_excluded], axis=1)
        test_scaled_df = pd.concat([test_scaled_df, test_excluded], axis=1)

        return train_scaled_df, test_scaled_df, scaler

    @staticmethod
    def robust_scaler(train_df: pd.DataFrame, test_df: pd.DataFrame, cols_to_scale: list[str], excluded_cols: list[str]) -> tuple[pd.DataFrame, pd.DataFrame, RobustScaler]:
        """
        Escala los datos utilizando estadísticas robustas frente a outliers (mediana y IQR).

        ➤ Usar cuando: los datos contienen outliers que podrían afectar a escaladores clásicos como StandardScaler o MinMaxScaler.

        :param train_df: DataFrame de entrenamiento.
        :param test_df: DataFrame de prueba.
        :param cols_to_scale: Columnas a escalar.
        :param excluded_cols: Columnas a mantener sin escalar.
        :return: Tupla (train_scaled_df, test_scaled_df, scaler).
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Copias y separación de columnas excluidas
        # -------------------------------------------------------------------------------------------------
        train_df: pd.DataFrame = train_df.copy()
        test_df: pd.DataFrame = test_df.copy()
        train_excluded = train_df[excluded_cols]
        test_excluded = test_df[excluded_cols]

        # -------------------------------------------------------------------------------------------------
        # -- 2: Aplicación del RobustScaler
        # -------------------------------------------------------------------------------------------------
        scaler: RobustScaler = RobustScaler()
        train_scaled = scaler.fit_transform(train_df[cols_to_scale])
        test_scaled = scaler.transform(test_df[cols_to_scale])

        train_scaled_df = pd.DataFrame(train_scaled, columns=cols_to_scale, index=train_df.index)
        test_scaled_df = pd.DataFrame(test_scaled, columns=cols_to_scale, index=test_df.index)

        # -------------------------------------------------------------------------------------------------
        # -- 3: Concatenar columnas excluidas
        # -------------------------------------------------------------------------------------------------
        train_scaled_df = pd.concat([train_scaled_df, train_excluded], axis=1)
        test_scaled_df = pd.concat([test_scaled_df, test_excluded], axis=1)

        return train_scaled_df, test_scaled_df, scaler

    @staticmethod
    def standard_scaler(train_df: pd.DataFrame, test_df: pd.DataFrame, cols_to_scale: list[str], excluded_cols: list[str]) -> tuple[pd.DataFrame, pd.DataFrame, StandardScaler]:
        """
        Escala los datos para que tengan media 0 y desviación estándar 1.

        ➤ Usar cuando: quieres normalizar variables continuas para algoritmos sensibles a la magnitud de los datos (p. ej. regresión, SVM).

        :param train_df: DataFrame de entrenamiento.
        :param test_df: DataFrame de prueba.
        :param cols_to_scale: Columnas a escalar.
        :param excluded_cols: Columnas a mantener sin escalar.
        :return: Tupla (train_scaled_df, test_scaled_df, scaler).
        """

        # -------------------------------------------------------------------------------------------------
        # -- 1: Copias y separación de columnas excluidas
        # -------------------------------------------------------------------------------------------------
        train_df: pd.DataFrame = train_df.copy()
        test_df: pd.DataFrame = test_df.copy()
        train_excluded = train_df[excluded_cols]
        test_excluded = test_df[excluded_cols]

        # -------------------------------------------------------------------------------------------------
        # -- 2: Aplicación del StandardScaler
        # -------------------------------------------------------------------------------------------------
        scaler: StandardScaler = StandardScaler()
        train_scaled = scaler.fit_transform(train_df[cols_to_scale])
        test_scaled = scaler.transform(test_df[cols_to_scale])

        train_scaled_df = pd.DataFrame(train_scaled, columns=cols_to_scale, index=train_df.index)
        test_scaled_df = pd.DataFrame(test_scaled, columns=cols_to_scale, index=test_df.index)

        # -------------------------------------------------------------------------------------------------
        # -- 3: Concatenar columnas excluidas
        # -------------------------------------------------------------------------------------------------
        train_scaled_df = pd.concat([train_scaled_df, train_excluded], axis=1)
        test_scaled_df = pd.concat([test_scaled_df, test_excluded], axis=1)

        return train_scaled_df, test_scaled_df, scaler

    # </editor-fold>



from hx_machine_learning_tools.ml_base_model import HxMachineLearningBaseModel
from _hx_model_evaluation_tools import EvaluateBinaryClassifier, BinaryClassifierMetricsCalculations
from _hx_model_evaluation_tools import MlShapBinaryAndRegressorTools
from xgboost import XGBClassifier
from typing import Dict, Literal, List, Any, Tuple
from joblib import dump
import pandas as pd
import numpy as np
import datetime
import os


class HxXtremeGradientBoostingBinaryClassifier(HxMachineLearningBaseModel):

    def __init__(self,
                 data_dict: Dict[str, pd.DataFrame],
                 hiperparams: dict,
                 metric: Literal['accuracy', 'precision', 'recall', 'f1', 'specificity', 'balanced_acc', 'roc_auc'],
                 problem_type: Literal['binary', 'multiclass'],
                 save_path: str | None = None,
                 bins: dict | None = None,
                 verbose: Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show',
                 random_state: int = 42):
        """
        Clase abstracta para construir modelos de clasificacion de Machine Learning
        :param data_dict: Diccionario que contiene el x_train, y_train, x_test, y_test
        :param save_path:
        :param hiperparams:
        :param bins:
        """
        super().__init__(data_dict, hiperparams, "XGB_binary_classifier", save_path)

        # --------------------------------------------------------------------------------------------
        # -- 0: Almaceno nombre de la clase y pinto la entrada
        # --------------------------------------------------------------------------------------------

        # ---- 0.1: Almaceno el nombre de la clase
        self.class_name: str = f"HxXtremeGradientBoostingBinaryClassifier"

        # ---- 0.2: Pinto la entrada
        self.CT.IT.intro_print(self.class_name)

        # ---- 0.3: Creo la propiedad model y la inicializo en None
        self.model = None

        # --------------------------------------------------------------------------------------------
        # -- 1: Asigno las propiedades abstractas
        # --------------------------------------------------------------------------------------------

        # ---- 1.1: Verbose
        self._verbose: Literal['show_all', 'show_basics', 'dont_show'] = verbose

        # ---- 1.2: Metric
        self._metric: Literal['accuracy', 'precision', 'recall', 'f1', 'specificity', 'balanced_acc', 'roc_auc'] = metric

        # ---- 1.3: Problem type
        self._problem_type: Literal['binary', 'multiclass'] = problem_type

        # --------------------------------------------------------------------------------------------
        # -- 2: Asigno las propiedades estandar
        # --------------------------------------------------------------------------------------------

        # ---- 2.1: Si bins es None, pongo el bins por defecto, sino lo asigno
        if bins is None:
            self.bins: Dict[str, List] = {
                "bins": [0, 0.125, 0.25, 0.375, 0.5,  0.625, 0.75, 0.875, 1],
                "labels": ["0-12.5", "12.5-25", "25-37.5", "37.5-50", "50-62.5", "62.5-75", "75-87.5", "87.5-100"]
            }
        else:
            self.bins = bins

        # ---- 2.2: Path donde se va a guardar la informacion del modelo
        self.model_save_path: str = self.get_save_path()

        # ---- 2.3: Random state con el que se van a inicializar los pesos, por defecto 42
        self.random_state: int = random_state

        # --------------------------------------------------------------------------------------------
        # -- 3: Reformulo el problem_type porque XGB lo requiere
        # --------------------------------------------------------------------------------------------

        # ---- 3.1: En caso de que el problem_type sea binary, reformateo a binary:logistic
        if self.problem_type == 'binary':
            self.problem_type = 'binary:logistic'

        # ---- 3.2: En caso de que el problem_type sea multiclass, reformateo a multi:softmax
        if self.problem_type == 'multiclass':
            self.problem_type = 'multi:softmax'


    # <editor-fold desc="Implementacion de metodos abstractos   ------------------------------------------------------------------------------------------------------------------">

    def get_save_path(self) -> str:
        """
        Metodo que hace lo siguiente:
        1: Busca dentro del path definido en la clase padre {save_path}/training_results/&Y_%m_%d/v_xxx si existe algúna carpeta llamda {self.model_name
        2: Si ya existe crea una nueva version, es decir {save_path}/training_results/&Y_%m_%d/v_xxx+1/{self.model_name y la asigna como self.save_path
        3: Si no existe, la asigna como self.save_path, es decir: {save_path}/training_results/&Y_%m_%d/v_xxx/{self.model_name
        :return: el save_path
        """

        # -- Si el path no existe, simplemente lo creo y lo asigno
        if not os.path.exists(f"{self.save_path}/{self.current_train_version}/{self.model_name}"):
            self.CT.OT.create_folder_if_not_exists(f"{self.save_path}/{self.current_train_version}/{self.model_name}")

        # -- Si el path existe, creo una nueva version y lo asigno
        else:
            self.current_train_version: str = self.new_training_version()

            self.CT.OT.create_folder_if_not_exists(f"{self.save_path}/{self.current_train_version}")
            self.CT.OT.create_folder_if_not_exists(f"{self.save_path}/{self.current_train_version}/{self.model_name}")

        # -- Actualizo las propiedades pertinentres de la clase padre
        self.day_train_versions: List[str] = self.list_versions()

        # -- Retorno el path
        return f"{self.save_path}/{self.current_train_version}/{self.model_name}"

    def fit_and_get_metrics(self, save_metrics_dict: bool) -> Dict[str, Any]:
        """
        Implementacion del metodo que entrena y devuelve el diccionario de métricas
        :param save_metrics_dict: Boleana que indica si quieres guardar el diccionario de metricas
        :return: diccionario de mettricas
        """
        # --------------------------------------------------------------------------------------------
        # -- 0: Pinto la entrada
        # --------------------------------------------------------------------------------------------
        start_date: datetime.datetime = datetime.datetime.now()
        self.CT.IT.sub_intro_print(f"{self.class_name}.fit_and_get_metrics: Inicio de entrenamiento a las {start_date}...")

        # --------------------------------------------------------------------------------------------
        # -- 1: Instancio el modelo y entreno
        # --------------------------------------------------------------------------------------------

        # ---- 1.1: Instancio el modelo y le asigno los patrámetros
        self.model: XGBClassifier = XGBClassifier(objective=self.problem_type, random_state=self.random_state, verbose=self.verbose, **self.hiperparams)

        # ---- 1.2: Realizo el entrenamiento
        self.model.fit(X=self.x_train, y=self.y_train.ravel())

        # ---- 1.3: Pinto el fin del entrenamiento y lo que ha tardado
        self.CT.IT.info_print(f"Entrenamiento completado a las {datetime.datetime.now()}. Ha tardado {datetime.datetime.now() - start_date}", "light_magenta")

        # --------------------------------------------------------------------------------------------
        # -- 2: Realizo las predicciones, mando a evaluar y retorno el diccionario de métricas
        # --------------------------------------------------------------------------------------------

        # ---- 2.1: Realizo las predicciones
        y_pred_train = self.model.predict(self.x_train_df)
        y_pred_test = self.model.predict(self.x_test_df)

        # ---- 2.2. Evalúo y almaceno las metricas
        metrics: Dict[str, Any] = EvaluateBinaryClassifier({"y_train": self.y_train,
                                                      "y_eval": None,
                                                      "y_test": self.y_test,
                                                      "y_pred_train": y_pred_train,
                                                      "y_pred_eval": None,
                                                      "y_pred_test": y_pred_test,
                                                      "target_col_name": self.target_col_name},
                                                           self.model_name,
                                                           self.metric).calculate_and_print_metrics()

        # ---- 2.3: Almaceno el diccionario de metricas como json en el path
        if save_metrics_dict:
            self.save_metrics_dict_json(metrics, self.model_save_path)

        # ---- 2.4. Pinto y retorno las métricas
        end_date = datetime.datetime.now()
        self.CT.IT.info_print(f"fit_and_get_metrics terminado a las {end_date}. Ha tardado {end_date - start_date}", "light_magenta")
        return metrics

    def fit_and_get_model_and_results(self) -> Tuple[Any, np.ndarray, np.ndarray, Dict[str, Any], List]:
        """
        Implementacion del metodo que entrena y devuelve (model, x_test_probs, x_train_probs, self.master_dict, model_weights)

        """
        # --------------------------------------------------------------------------------------------
        # -- 0: Pinto la entrada
        # --------------------------------------------------------------------------------------------
        start_date: datetime.datetime = datetime.datetime.now()
        self.CT.IT.sub_intro_print(f"{self.class_name}.fit_and_get_model_and_results: Inicio de entrenamiento a las {start_date}...")

        # --------------------------------------------------------------------------------------------
        # -- 1: Instancio el modelo y entreno
        # --------------------------------------------------------------------------------------------

        # ---- 1.1: Instancio el modelo y le asigno los patrámetros
        self.model: XGBClassifier = XGBClassifier(objective=self.problem_type, random_state=self.random_state, **self.hiperparams)

        # ---- 1.2: Realizo el entrenamiento
        self.model.fit(X=self.x_train, y=self.y_train.ravel(), verbose=self.verbose)

        # ---- 1.3: Almaceno el modelo y el modulo del que proviene
        model_class_name: str = self.model.__class__.__name__
        model_module_name: str = self.model.__class__.__module__

        # ---- 1.4: Pinto el fin del entrenamiento y lo que ha tardado
        self.CT.IT.info_print(f"Entrenamiento completado a las {datetime.datetime.now()}. Ha tardado {datetime.datetime.now() - start_date}", "light_magenta")

        # --------------------------------------------------------------------------------------------
        # -- 2: Realizo las predicciones, mando a evaluar y retorno el diccionario de métricas
        # --------------------------------------------------------------------------------------------

        # ---- 2.1: Realizo las predicciones
        y_pred_train = self.model.predict(self.x_train_df)
        y_pred_test = self.model.predict(self.x_test_df)

        # ---- 2.2. Evalúo las metricas
        EvaluateBinaryClassifier(
            {
                "y_train": self.y_train,
                "y_eval": None,
                "y_test": self.y_test,
                "y_pred_train": y_pred_train,
                "y_pred_eval": None,
                "y_pred_test": y_pred_test,
                "target_col_name": self.target_col_name
            },
            self.model_name,
            self.metric).calculate_and_print_metrics()

        # --------------------------------------------------------------------------------------------
        # -- 3: Almaceno el modelo en su path con joblib.dump y obtengo los pesos
        # --------------------------------------------------------------------------------------------

        # ---- 3.1: Almacenamiento del modelo
        dump(self.model, f"{self.model_save_path}/{self.class_name}.joblib")

        # ---- 3.2: Obtencion de pesos (llamo al metodo polimorfeado self.get_weights)
        model_weights: List = self.get_weights(self.model)

        # ---- 3.3: Creacion de diccionario en el que asocio cada peso con su columna
        model_weights_dict: Dict[str, Any] = {col: weight for col, weight in zip([z for z in self.x_train_df.columns], [float(x) for x in model_weights[0]])}

        # --------------------------------------------------------------------------------------------
        # -- 4: Realizo las predicciones de probabilidad en test (valor continuo entre 0 y 1) y las almaceno
        # --------------------------------------------------------------------------------------------

        # ---- 4.1: Obtengo una matriz con la probabilidad negativa y positiva de x_train y x_test
        x_train_probs: np.ndarray[tuple[int]] = self.model.predict_proba(self.x_train_df)
        x_test_probs: np.ndarray[tuple[int]] = self.model.predict_proba(self.x_test_df)

        # ---- 4.2: Creo un diccionario llamado probs_result_df en el que almaceno el y_test real y los de las probs
        probs_result_dict: Dict[str, Any] = {
            f'real_value_{self.target_col_name}': self.y_test[:, 0],
            f'negative_proba_{self.target_col_name}': x_test_probs[:, 0],
            f'positive_proba_{self.target_col_name}': x_test_probs[:, 1]
        }

        # ---- 4.3: Asigno probs_result_df a un dataframe para trabajarlo con mas facilidad
        probs_result_df: pd.DataFrame = pd.DataFrame(probs_result_dict)

        # --------------------------------------------------------------------------------------------
        # -- 5: Instancio BinaryClassifierMetricsCalculations para obtener toda la info del resultado y rellenar el self.master_dict
        # --------------------------------------------------------------------------------------------

        # ---- 5.1: Obtengo el diccionario de metricas que proporciona BinaryClassifierMetricsCalculations.run
        metrics_dict: dict = BinaryClassifierMetricsCalculations(
            probs_result_df,
            self.model_save_path,
            self.x_test_df,
            self.model,
            model_weights,
            self.target_col_name,
            self.model_name,
            self.bins
        ).run()

        # ---- 5.2: Relleno el self.master_dict
        self.master_result_dict[f"{self.model_name}"]: dict = {}
        self.master_result_dict[f"{self.model_name}"]["columns"] = [z for z in self.x_train_df.columns]
        self.master_result_dict[f"{self.model_name}"]["target"] = self.target_col_name
        self.master_result_dict[f"{self.model_name}"]["best_params"] = self.hiperparams
        self.master_result_dict[f"{self.model_name}"]["metrics"] = metrics_dict
        self.master_result_dict[f"{self.model_name}"]["model_weights"] = model_weights_dict
        self.master_result_dict[f"{self.model_name}"]["model_class_name"] = model_class_name
        self.master_result_dict[f"{self.model_name}"]["model_module_name"] = model_module_name
        self.master_result_dict[f"{self.model_name}"]["model_version"] = f"{self.today_str_daye}_{self.current_train_version}"

        # --------------------------------------------------------------------------------------------
        # -- 6: Pinto y retorno la tupla (model, x_test_probs, x_train_probs, self.master_dict, model_weights)
        # --------------------------------------------------------------------------------------------

        # ---- 6.1: Pinto el final
        end_date = datetime.datetime.now()
        self.CT.IT.info_print(f"fit_and_get_model_and_results terminado a las {end_date}. Ha tardado {end_date - start_date}", "light_magenta")

        # ---- 6.2: Almaceno el self.master_dict
        self.save_metrics_dict_json(self.master_result_dict, self.model_save_path, "complete_train_result")

        # ---- 6.3: Retorno la tupla
        return self.model, x_test_probs, x_train_probs, self.master_result_dict, model_weights

    def get_weights(self, model) -> List:
        """
        Implementacion del metodo que devuelve una lista de ndarrays con los pesos del modelo

        """

        return [model.feature_importances_] if hasattr(model, "feature_importances_") else [None]

    def get_model_save_path(self) -> str:
        return self.model_save_path

    def execute_shap_analysis(self, sample: bool = True, num_features_to_show: int = 100, num_sample: int = 500) -> pd.DataFrame:
        """
        Metodo para ejecutar el analisis SHAP
        :return:
        """

        return MlShapBinaryAndRegressorTools(self.x_test_df, self.model_name, self.model_save_path, self.model, sample, num_features_to_show, num_sample).run()

    # </editor-fold>

    # <editor-fold desc="Implementacion de propiedades abstractas   --------------------------------------------------------------------------------------------------------------">

    @property
    def verbose(self):
        verbose_dict: dict = {'show_all': 1, 'dont_show': 0}
        return verbose_dict[self._verbose]

    @property
    def metric(self):
        if self._metric not in ['accuracy', 'precision', 'recall', 'f1', 'specificity', 'balanced_acc', 'roc_auc']:
            raise ValueError(f'Invalid metric: {self._metric}')
        return self._metric

    @property
    def problem_type(self):
        if self._problem_type not in ['binary', 'multiclass', 'binary:logistic', 'multi:softmax']:
            raise ValueError(f'Invalid problem_type: {self._problem_type}. Must be binary or multiclass')
        return self._problem_type

    # </editor-fold>
    @problem_type.setter
    def problem_type(self, value):
        self._problem_type = value










from _hx_model_evaluation_tools.hx_model_evaluation_and_metrics.binary_classifiers import EvaluateBinaryClassifier, BinaryClassifierMetricsCalculations
from _hx_model_evaluation_tools.hx_model_evaluation_and_metrics.multiclass_classifiers import EvaluateMulticlassClassifier, MulticlassClassifierMetricsCalculations
from _hx_model_evaluation_tools.hx_model_evaluation_and_metrics.regressors import EvaluateRegressor, RegressorMetricsCalculations


from _hx_model_evaluation_tools.hx_feature_importances import HxFeatureImportancesandCorrelations
from _hx_model_evaluation_tools.hx_model_evaluation_and_metrics import (EvaluateRegressor, RegressorMetricsCalculations,
                                                                        EvaluateBinaryClassifier, BinaryClassifierMetricsCalculations,
                                                                        EvaluateMulticlassClassifier, MulticlassClassifierMetricsCalculations)
from _hx_model_evaluation_tools.hx_model_evaluation_graphics import HxModelEvaluationGraphics
from _hx_model_evaluation_tools.hx_shap_tools import MlShapBinaryAndRegressorTools, MlShapToolsMulticlass, DlShapToolsBinaryRegressor, DlShapToolsMulticlass


from _hx_model_evaluation_tools.hx_shap_tools.dl_shap_tools import DlShapToolsBinaryRegressor, DlShapToolsMulticlass
from _hx_model_evaluation_tools.hx_shap_tools.ml_shap_tools import MlShapBinaryAndRegressorTools, MlShapToolsMulticlass


from setuptools import setup, find_packages

setup(
    name="hx_ml_dl_tools",
    version="1.1.2",
    author="Daniel Sarabia Torres aka Huexmaister",
    author_email="dsarabiatorres@gmail.com",
    description="Librería desarrollo de modelos de ML y DL",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Huexmaister/hx_ml_dl_tools",
    packages=find_packages(),  # Busca automáticamente todos los paquetes
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "catboost==1.2.8",
        "category_encoders==2.8.1",
        "constants_and_tools>=2.2.2",
        "imblearn==0.0",
        "joblib==1.5.2",
        "lightgbm==4.6.0",
        "matplotlib==3.10.6",
        "openpyxl==3.1.5",
        "scikit-learn==1.7.2",
        "shap==0.48.0",
        "tensorflow==2.20.0",
        "xgboost==3.0.5",
    ],
    python_requires=">=3.10",
)



from hx_ml_dl_tools_examples.regression_examples import DiabetesExample, CaliforniaHousingExample
from hx_ml_dl_tools_examples.binary_classification_examples import BreastCancerExample, CreditFraudExample
from hx_ml_dl_tools_examples.multiclass_classification_examples import IrisMulticlassExample, DigitsMulticlassExample, WineQualityMulticlassExample


from hx_ml_dl_tools import HxLightGbmBinaryClassifier, HxXtremeGradientBoostingBinaryClassifier, HxCatBoostBinaryClassifier
from hx_ml_dl_tools import HxDenseNeuralNetworkBinaryClassifier
from sklearn.preprocessing import StandardScaler, LabelEncoder
from sklearn.datasets import load_breast_cancer, fetch_openml
from sklearn.model_selection import train_test_split
from hx_hyperparameters_builder import *
from info_tools import InfoTools
import pandas as pd
import numpy as np


class BreastCancerExample:
    def __init__(self):
        """
        Ejemplo de clasificacion binaria de modelos de ML y DL para el ejemplo de la diabetes de sklearn
        """

        # --------------------------------------------------------------------------------------
        # -- 0: Instancio InfoTools y pinto la entrada
        # --------------------------------------------------------------------------------------

        # ---- 0.1: Instancia de InfoTools
        self.IT: InfoTools = InfoTools()

        # ---- 0.2: Pinto la entrada
        self.IT.header_print(f"DiabetesExample: Ejemplo de clasificacion con sklearn.datasets.breast_cancer utilizando hx_ml_dl_tools")

        # --------------------------------------------------------------------------------------
        # -- 1: Descargo el dataset y obtengo el data dict
        # --------------------------------------------------------------------------------------

        # ---- 1.1: Descargo el dataset
        self.cancer_dataset = load_breast_cancer()

        # ---- 1.2: Transformo a dataframe (Lo requieren los modelos)
        self.x = pd.DataFrame(self.cancer_dataset.data, columns=self.cancer_dataset.feature_names)
        self.y = pd.DataFrame(self.cancer_dataset.target, columns=["target"])

        # ---- 1.3: Divido en train y test
        self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(self.x, self.y, test_size=0.2, random_state=42)

        # ---- 1.4: Creo el diccionario en el formato requerido por los modelos
        self.data_dict = {"x_train": self.x_train, "y_train": self.y_train, "x_test": self.x_test, "y_test": self.y_test}

        # --------------------------------------------------------------------------------------
        # -- 2: Obtengo informacion basica del dataset
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Información del dataset
        self.IT.sub_intro_print(f"Dataset BreastCancer cargado:")
        self.IT.info_print(f"Muestras totales: {self.x.shape[0]}")
        self.IT.info_print(f"Características: {self.x.shape[1]}")
        self.IT.info_print(f"Rango target: {self.y['target'].min():.2f} - {self.y['target'].max():.2f}")
        self.IT.info_print(f"Train samples: {self.x_train.shape[0]}")
        self.IT.info_print(f"Test samples: {self.x_test.shape[0]}")

        # ---- 2.2: Informacion de distribucion
        self.get_dataset_info()

    def execute_example(self):
        """
        Metodo que ejecuta en cadena los metodos de entrenamiento
        :return:
        """

        self._light_gbm_classifier()
        self._xg_boost_classifier()
        self._cat_boost_classifier()
        self._dnn_binary_calssifier()

    def get_dataset_info(self):
        """
        Metodo adicional para obtener información detallada del dataset
        """
        self.IT.sub_intro_print("Estadísticas del target:")

        self.IT.info_print(f"Media: {self.y['target'].mean():.2f}")
        self.IT.info_print(f"Mediana: {self.y['target'].median():.2f}")
        self.IT.info_print(f"Std: {self.y['target'].std():.2f}")
        self.IT.info_print(f"Min: {self.y['target'].min():.2f}")
        self.IT.info_print(f"Max: {self.y['target'].max():.2f}")

    # <editor-fold desc="Modelos   -----------------------------------------------------------------------------------------------------------------------------------------------">

    def _dnn_binary_calssifier(self):
        """
        Metodo para realizar el entrenamiento de la red neuronal densa de clasificacion binaria
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar
        # --------------------------------------------------------------------------------------

        # ---- 1.1: Opcion de pasarselos a mano
        # dnn_hyperparams: Dict[str, Union[int, float]] = {'units': 8, 'dropout': 0.5, 'learning_rate': 0.01, 'regularization': 0.1,  'epochs': 200}

        # ---- 1.2: Opcion de usar la clase HxDnnHyperparameterBuilder

        # ------ 1.2.1: Obtener parámetros aleatorios dentro del rango preestablecido en HxDnnHyperparameterBuilder
        # dnn_hyperparams: Dict[str, Union[int, float]] = HxDnnHyperparameterBuilder().get_random_hyperparams()

        # ------ 1.2.2: Pasarle los hiperparametros al constructor para que valide si hay alguno que se usa poco o no existe (USAMOS ESTA PARA EL EJEMPLO)
        dnn_hyperparams = HxDnnHyperparameterBuilder({'units': 8, 'dropout': 0.5, 'learning_rate': 0.01, 'regularization': 0.1,  'epochs': 200}).validate_and_get_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase que me va a proporcionar un toolkit para ese modelo en concreto
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados (Entrar en la clase para ver detalle)
        dnn: HxDenseNeuralNetworkBinaryClassifier = HxDenseNeuralNetworkBinaryClassifier(
            self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            dnn_hyperparams,  # Diccionario de hiperparametros
            'accuracy',  # Literal['accuracy', 'precision', 'recall', 'f1', 'specificity', 'balanced_acc', 'roc_auc']
            'binary',  # Literal['binary', 'multiclass']
            1,  # Literal[0, 1, 2, 3] = 1 ({0: "<", 1: ">", 2: "<>", 3: "><"}) Arquitectura de la red
            1,  # Capas ocultas de la red (Toda red tiene 1 capa de entrada + N capas ocultas + 1 capa de salida)
            "relu",  # Literal['relu', 'tanh', 'sigmoid'] = 'relu' (Funcion de activacion)
            "l1",  # Literal['l1', 'l2'] = 'l2'  (Tipo de regularizacion que se aplicará en caso de aplicar regularizacion)
            "adam",  # Literal['sgd', 'adam', 'rmsprop'] = 'adam'  (Optimizador que se va a utilizar)
            True,  # bool (Incluir o no capas de dropout (en caso de no incluirse, el hiperparametro dropout no hace nada))
            True,  # bool (Incluir capa de batch normalization)
            15,  # Si en N epochs no ha mejorado la métrica, aplica un earlyStop callback y devuelve el mejor modelo hasta la fecha
            None,  # Path donde se almacenarán los resultados (por defecto None, ya que se crean internamente las estructuras de carpetas)
            None,  #  dict | None = None (Diccionario de estratificacion, en binary está por defecto entre 0 y 1)
            "dont_show",  # Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show' (Cantidad de informacion adicional que se muestra en consola)
            42,  # Random state de incializacion de pesos para garantizar reproducibilidad
        )

        # ---- 2.2: [Metodo]: Permite entrenar el modelo, pintar las metricas en pantalla y devolver un diccionario de las metricas
        # dnn.fit_and_get_metrics(save_metrics_dict=False)

        # ---- 2.3: [Metodo]: Metodo principal de la clase que ejecuta internamente 'fit_and_get_metrics', almacena el modelo junto con graficos e info adicional
        dnn.fit_and_get_model_and_results()

        # ---- 2.4: [Metodo]: Permite ejecutar el analisis SHAP del modelo
        dnn.execute_shap_analysis(
            sample=True,  # Realizar un sampleo de la data antes de calcular SHAP (!! Es muy intensivo en cómputo si no se samplea !!)
            num_features_to_show=100,  # En los gráficos se van a mostrar las N características mas importantes (con mas peso)
            num_sample=200,  # Cantidad de muestra (filas) con las que se va a realizar el analisis
            background_sample=100,  # Número de instancias para background (explicador deep/grad).
        )

    def _light_gbm_classifier(self):
        """
        Metodo para realizar el entrenamiento del lgbm de regresion
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar
        # --------------------------------------------------------------------------------------

        # ---- 1.1: En este caso, los obtengo aleatorios (en el ejemplo de la dnn están todas las opciones)
        lgbm_hyperparams: Dict[str, Union[int, float]] = HxLightgbmHyperparameterBuilder().get_random_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase que me va a proporcionar un toolkit para ese modelo en concreto
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados (Entrar en la clase para ver detalle)
        lgbm: HxLightGbmBinaryClassifier = HxLightGbmBinaryClassifier(
            data_dict=self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            hiperparams=lgbm_hyperparams,  # Diccionario con los hiperparametros del LGBM
            metric="roc_auc",  # Literal['accuracy', 'precision', 'recall', 'f1', 'specificity', 'balanced_acc', 'roc_auc'], Metrica que vamos a evaluar
            problem_type="binary",  # Literal['binary', 'multiclass']
            save_path=None,  # Path donde se almacenarán los resultados (por defecto None, ya que se crean internamente las estructuras de carpetas)
            bins=None,  # En binary viene por defecto
            verbose="dont_show",  # Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show' (Cantidad de informacion adicional que se muestra en consola)
            random_state=42,  # Random state de incializacion de pesos para garantizar reproducibilidad

        )

        # ---- 2.2: [Metodo]: Permite entrenar el modelo, pintar las metricas en pantalla y devolver un diccionario de las metricas
        # lgbm.fit_and_get_metrics(save_metrics_dict=False)

        # ---- 2.3: [Metodo]: Metodo principal de la clase que ejecuta internamente 'fit_and_get_metrics', almacena el modelo junto con graficos e info adicional
        lgbm.fit_and_get_model_and_results()

        # ---- 2.4: [Metodo]: Permite ejecutar el analisis SHAP del modelo
        lgbm.execute_shap_analysis(
            sample=True,  # Realizar un sampleo de la data antes de calcular SHAP (!! Es muy intensivo en cómputo si no se samplea !!)
            num_features_to_show=100,  # En los gráficos se van a mostrar las N características mas importantes (con mas peso)
            num_sample=200,  # Número de instancias para background (explicador deep/grad).
        )

    def _xg_boost_classifier(self):
        """
        Metodo para realizar el entrenamiento del xgboost de clasificacion
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar
        # --------------------------------------------------------------------------------------

        # ---- 1.1: En este caso, los obtengo aleatorios (en el ejemplo de la dnn están todas las opciones)
        xgboost_hyperparams: Dict[str, Union[int, float]] = HxXgboostHyperparameterBuilder().get_random_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase que me va a proporcionar un toolkit para ese modelo en concreto
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados (Entrar en la clase para ver detalle)
        xgboost: HxXtremeGradientBoostingBinaryClassifier = HxXtremeGradientBoostingBinaryClassifier(
            data_dict=self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            hiperparams=xgboost_hyperparams,  # Diccionario con los hiperparametros del XGB
            metric="roc_auc",  # Literal['accuracy', 'precision', 'recall', 'f1', 'specificity', 'balanced_acc', 'roc_auc'], Metrica que vamos a evaluar
            problem_type="binary",  # Literal['binary', 'multiclass']
            save_path=None,  # Path donde se almacenarán los resultados (por defecto None, ya que se crean internamente las estructuras de carpetas)
            bins=None,  # En binary viene por defecto
            verbose="dont_show",  # Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show' (Cantidad de informacion adicional que se muestra en consola)
            random_state=42,  # Random state de incializacion de pesos para garantizar reproducibilidad

        )

        # ---- 2.2: [Metodo]: Permite entrenar el modelo, pintar las metricas en pantalla y devolver un diccionario de las metricas
        # xgboost.fit_and_get_metrics(save_metrics_dict=False)

        # ---- 2.3: [Metodo]: Metodo principal de la clase que ejecuta internamente 'fit_and_get_metrics', almacena el modelo junto con graficos e info adicional
        xgboost.fit_and_get_model_and_results()

        # ---- 2.4: [Metodo]: Permite ejecutar el analisis SHAP del modelo
        xgboost.execute_shap_analysis(
            sample=True,  # Realizar un sampleo de la data antes de calcular SHAP (!! Es muy intensivo en cómputo si no se samplea !!)
            num_features_to_show=100,  # En los gráficos se van a mostrar las N características mas importantes (con mas peso)
            num_sample=200,  # Número de instancias para background (explicador deep/grad).
        )

    def _cat_boost_classifier(self):
        """
        Metodo para realizar el entrenamiento del catboost de regresion
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar
        # --------------------------------------------------------------------------------------

        # ---- 1.1: En este caso, los obtengo aleatorios (en el ejemplo de la dnn están todas las opciones)
        catboost_hyperparams: Dict[str, Union[int, float]] = HxCatBoostHyperparameterBuilder().get_random_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase que me va a proporcionar un toolkit para ese modelo en concreto
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados (Entrar en la clase para ver detalle)
        catboost: HxCatBoostBinaryClassifier = HxCatBoostBinaryClassifier(
            data_dict=self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            hiperparams=catboost_hyperparams,  # Diccionario con los hiperparametros del XGB
            metric="roc_auc",  # Literal['accuracy', 'precision', 'recall', 'f1', 'specificity', 'balanced_acc', 'roc_auc'], Metrica que vamos a evaluar
            problem_type="binary",  # Literal['binary', 'multiclass']
            save_path=None,  # Path donde se almacenarán los resultados (por defecto None, ya que se crean internamente las estructuras de carpetas)
            bins=None,  # En binary viene por defecto
            verbose="dont_show",  # Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show' (Cantidad de informacion adicional que se muestra en consola)
            random_state=42,  # Random state de incializacion de pesos para garantizar reproducibilidad

        )

        # ---- 2.2: [Metodo]: Permite entrenar el modelo, pintar las metricas en pantalla y devolver un diccionario de las metricas
        # catboost.fit_and_get_metrics(save_metrics_dict=False)

        # ---- 2.3: [Metodo]: Metodo principal de la clase que ejecuta internamente 'fit_and_get_metrics', almacena el modelo junto con graficos e info adicional
        catboost.fit_and_get_model_and_results()

        # ---- 2.4: [Metodo]: Permite ejecutar el analisis SHAP del modelo
        catboost.execute_shap_analysis(
            sample=True,  # Realizar un sampleo de la data antes de calcular SHAP (!! Es muy intensivo en cómputo si no se samplea !!)
            num_features_to_show=100,  # En los gráficos se van a mostrar las N características mas importantes (con mas peso)
            num_sample=200,  # Número de instancias para background (explicador deep/grad).
        )

    # </editor-fold>


class CreditFraudExample:
    def __init__(self, sample_size: int = 100000):
        """
        Ejemplo de clasificación binaria con dataset de Fraud Detection (más grande y profesional)
        Dataset: Credit Card Fraud Detection from Kaggle (via OpenML)
        """

        # --------------------------------------------------------------------------------------
        # -- 0: Instancio InfoTools y pinto la entrada
        # --------------------------------------------------------------------------------------

        self.IT: InfoTools = InfoTools()
        self.IT.header_print(f"CreditFraudExample: Clasificación binaria con dataset profesional de Fraud Detection")

        # --------------------------------------------------------------------------------------
        # -- 1: Descargo el dataset de Fraud Detection (más grande y realista)
        # --------------------------------------------------------------------------------------

        self.IT.info_print("Cargando dataset de Credit Card Fraud Detection...")

        try:
            # Dataset de fraud detection (más de 280,000 transacciones)
            fraud_data = fetch_openml(name='creditcard', version=1, as_frame=True)
            self.x = fraud_data.data
            self.y = fraud_data.target

            # Si el target es string, convertirlo a numérico
            if self.y.dtype == 'object':
                self.y = self.y.astype(int)

        except Exception as e:
            self.IT.info_print(f"Error cargando dataset de fraud: {e}")
            self.IT.info_print("Cargando dataset alternativo Adult Income...")
            # Fallback a Adult Income dataset
            adult_data = fetch_openml(name='adult', version=2, as_frame=True)
            self.x = adult_data.data
            self.y = adult_data.target
            # Codificar target binario
            le = LabelEncoder()
            self.y = pd.Series(le.fit_transform(self.y))

        # --------------------------------------------------------------------------------------
        # -- 2: Preprocesamiento para dataset grande - CORRECCIÓN CRÍTICA
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Asegurar que x e y tengan el mismo índice
        self.x = self.x.reset_index(drop=True)
        self.y = self.y.reset_index(drop=True)

        # ---- 2.2: Samplear si el dataset es muy grande para pruebas rápidas
        if len(self.x) > sample_size:
            self.IT.info_print(f"Realizando sampleo de {sample_size} muestras...")
            indices = np.random.choice(len(self.x), sample_size, replace=False)
            self.x = self.x.iloc[indices].reset_index(drop=True)
            self.y = self.y.iloc[indices].reset_index(drop=True)

        # ---- 2.3: Limpiar y preprocesar
        self.x = self._clean_dataframe(self.x)

        # ---- 2.4: Escalar características (CRÍTICO para redes neuronales)
        scaler = StandardScaler()
        self.x_scaled = pd.DataFrame(scaler.fit_transform(self.x), columns=self.x.columns)

        # ---- 2.5: Convertir y a DataFrame - CORRECCIÓN: asegurar mismo índice
        self.y_df = pd.DataFrame(self.y.values, columns=["target"])  # Usar .values en lugar de la serie directamente
        self.y_df.index = self.x_scaled.index  # Asegurar mismos índices

        # ---- 2.6: Verificar que las dimensiones coincidan
        self.IT.info_print(f"Verificando dimensiones: X_scaled {self.x_scaled.shape}, y_df {self.y_df.shape}")

        if len(self.x_scaled) != len(self.y_df):
            self.IT.info_print("Ajustando dimensiones...")
            # Tomar el mínimo común de muestras
            min_len = min(len(self.x_scaled), len(self.y_df))
            self.x_scaled = self.x_scaled.iloc[:min_len]
            self.y_df = self.y_df.iloc[:min_len]

        # ---- 2.7: Dividir en train y test con estratificación
        self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(
            self.x_scaled, self.y_df, test_size=0.2, random_state=42, stratify=self.y_df
        )

        # ---- 2.8: Crear data_dict
        self.data_dict = {
            "x_train": self.x_train,
            "y_train": self.y_train,
            "x_test": self.x_test,
            "y_test": self.y_test
        }

        # --------------------------------------------------------------------------------------
        # -- 3: Información del dataset
        # --------------------------------------------------------------------------------------

        self.IT.sub_intro_print(f"Dataset Profesional Cargado:")
        self.IT.info_print(f"Muestras totales: {self.x.shape[0]:,}")
        self.IT.info_print(f"Características: {self.x.shape[1]}")

        class_info = self.y_df['target'].value_counts()
        total = len(self.y_df)
        self.IT.info_print(f"Clases: {class_info.to_dict()}")
        self.IT.info_print(f"Proporciones: {({k: f'{v / total * 100:.1f}%' for k, v in class_info.items()})}")

        self.IT.info_print(f"Train samples: {self.x_train.shape[0]:,}")
        self.IT.info_print(f"Test samples: {self.x_test.shape[0]:,}")

        self.get_dataset_info()

    @staticmethod
    def _clean_dataframe(df: pd.DataFrame) -> pd.DataFrame:
        """Limpieza básica del dataframe"""
        # Eliminar columnas con demasiados missing values
        df = df.dropna(axis=1, thresh=len(df) * 0.8)

        # Llenar valores missing con la mediana para numéricos y moda para categóricos
        for col in df.columns:
            if df[col].dtype in ['float64', 'int64']:
                df[col] = df[col].fillna(df[col].median())
            else:
                df[col] = df[col].fillna(df[col].mode()[0] if len(df[col].mode()) > 0 else 'Unknown')

        # Convertir variables categóricas a numéricas si las hay
        categorical_cols = df.select_dtypes(include=['object', 'category']).columns
        if len(categorical_cols) > 0:
            le = LabelEncoder()
            for col in categorical_cols:
                df[col] = le.fit_transform(df[col].astype(str))

        return df

    def execute_example(self):
        """
        Método que ejecuta en cadena los métodos de entrenamiento
        """
        self.IT.sub_intro_print("Iniciando entrenamiento de modelos para Fraud Detection")

        # Ejecutar modelos (comenta los que no quieras ejecutar)
        self._light_gbm_classifier()
        self._xg_boost_classifier()
        self._cat_boost_classifier()
        self._dnn_binary_classifier()

        self.IT.sub_intro_print("Entrenamiento completado para todos los modelos")

    def get_dataset_info(self):
        """Información detallada del dataset"""
        self.IT.sub_intro_print("Análisis Detallado del Dataset:")

        class_counts = self.y_df['target'].value_counts()
        class_percentages = (class_counts / len(self.y_df)) * 100

        self.IT.info_print("Distribución de clases:")
        for class_val, count in class_counts.items():
            percentage = class_percentages[class_val]
            class_name = "Fraude" if class_val == 1 else "Normal" if 'creditcard' in str(self.x.columns) else "Clase 0/1"
            self.IT.info_print(f"  - {class_name} (Clase {class_val}): {count:,} muestras ({percentage:.2f}%)")

        self.IT.info_print(f"Balance: {class_percentages[0]:.2f}% vs {class_percentages[1]:.2f}%")
        self.IT.info_print(f"Dimensiones: {self.x.shape[0]:,} filas × {self.x.shape[1]} columnas")

    # <editor-fold desc="Modelos -------------------------------------------------------------------------------------------------------------------------------------------------">

    def _dnn_binary_classifier(self):
        """
        Metodo para realizar el entrenamiento de la red neuronal densa de clasificacion binaria
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar
        # --------------------------------------------------------------------------------------

        # ---- 1.1: Opcion de pasarselos a mano
        # dnn_hyperparams: Dict[str, Union[int, float]] = {'units': 8, 'dropout': 0.5, 'learning_rate': 0.01, 'regularization': 0.1,  'epochs': 200}

        # ---- 1.2: Opcion de usar la clase HxDnnHyperparameterBuilder

        # ------ 1.2.1: Obtener parámetros aleatorios dentro del rango preestablecido en HxDnnHyperparameterBuilder
        # dnn_hyperparams: Dict[str, Union[int, float]] = HxDnnHyperparameterBuilder().get_random_hyperparams()

        # ------ 1.2.2: Pasarle los hiperparametros al constructor para que valide si hay alguno que se usa poco o no existe (USAMOS ESTA PARA EL EJEMPLO)
        dnn_hyperparams = HxDnnHyperparameterBuilder({
            'units': 128,  # Más unidades para dataset grande
            'dropout': 0.3,  # Dropout moderado
            'learning_rate': 0.001,  # Learning rate estándar
            'regularization': 0.001,  # Regularización ligera
            'epochs': 50  # Menos épocas para dataset grande
        }).validate_and_get_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase que me va a proporcionar un toolkit para ese modelo en concreto
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados (Entrar en la clase para ver detalle)
        dnn: HxDenseNeuralNetworkBinaryClassifier = HxDenseNeuralNetworkBinaryClassifier(
            self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            dnn_hyperparams,  # Diccionario de hiperparametros
            'accuracy',  # Literal['accuracy', 'precision', 'recall', 'f1', 'specificity', 'balanced_acc', 'roc_auc']
            'binary',  # Literal['binary', 'multiclass']
            1,  # Literal[0, 1, 2, 3] = 1 ({0: "<", 1: ">", 2: "<>", 3: "><"}) Arquitectura de la red
            1,  # Capas ocultas de la red (Toda red tiene 1 capa de entrada + N capas ocultas + 1 capa de salida)
            "relu",  # Literal['relu', 'tanh', 'sigmoid'] = 'relu' (Funcion de activacion)
            "l2",  # Literal['l1', 'l2'] = 'l2'  (Tipo de regularizacion que se aplicará en caso de aplicar regularizacion)
            "adam",  # Literal['sgd', 'adam', 'rmsprop'] = 'adam'  (Optimizador que se va a utilizar)
            True,  # bool (Incluir o no capas de dropout (en caso de no incluirse, el hiperparametro dropout no hace nada))
            True,  # bool (Incluir capa de batch normalization)
            15,  # Si en N epochs no ha mejorado la métrica, aplica un earlyStop callback y devuelve el mejor modelo hasta la fecha
            None,  # Path donde se almacenarán los resultados (por defecto None, ya que se crean internamente las estructuras de carpetas)
            None,  #  dict | None = None (Diccionario de estratificacion, en binary está por defecto entre 0 y 1)
            "dont_show",  # Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show' (Cantidad de informacion adicional que se muestra en consola)
            42,  # Random state de incializacion de pesos para garantizar reproducibilidad
        )

        # ---- 2.2: [Metodo]: Permite entrenar el modelo, pintar las metricas en pantalla y devolver un diccionario de las metricas
        # dnn.fit_and_get_metrics(save_metrics_dict=False)

        # ---- 2.3: [Metodo]: Metodo principal de la clase que ejecuta internamente 'fit_and_get_metrics', almacena el modelo junto con graficos e info adicional
        dnn.fit_and_get_model_and_results()

        # ---- 2.4: [Metodo]: Permite ejecutar el analisis SHAP del modelo
        dnn.execute_shap_analysis(
            sample=True,  # Realizar un sampleo de la data antes de calcular SHAP (!! Es muy intensivo en cómputo si no se samplea !!)
            num_features_to_show=100,  # En los gráficos se van a mostrar las N características mas importantes (con mas peso)
            num_sample=50,  # Cantidad de muestra (filas) con las que se va a realizar el analisis
            background_sample=50,  # Número de instancias para background (explicador deep/grad).
        )

    def _light_gbm_classifier(self):
        """
        Metodo para realizar el entrenamiento del lgbm de regresion
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar
        # --------------------------------------------------------------------------------------

        # ---- 1.1: En este caso, los obtengo aleatorios (en el ejemplo de la dnn están todas las opciones)
        lgbm_hyperparams: Dict[str, Union[int, float]] = HxLightgbmHyperparameterBuilder().get_random_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase que me va a proporcionar un toolkit para ese modelo en concreto
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados (Entrar en la clase para ver detalle)
        lgbm: HxLightGbmBinaryClassifier = HxLightGbmBinaryClassifier(
            data_dict=self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            hiperparams=lgbm_hyperparams,  # Diccionario con los hiperparametros del LGBM
            metric="roc_auc",  # Literal['accuracy', 'precision', 'recall', 'f1', 'specificity', 'balanced_acc', 'roc_auc'], Metrica que vamos a evaluar
            problem_type="binary",  # Literal['binary', 'multiclass']
            save_path=None,  # Path donde se almacenarán los resultados (por defecto None, ya que se crean internamente las estructuras de carpetas)
            bins=None,  # En binary viene por defecto
            verbose="dont_show",  # Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show' (Cantidad de informacion adicional que se muestra en consola)
            random_state=42,  # Random state de incializacion de pesos para garantizar reproducibilidad

        )

        # ---- 2.2: [Metodo]: Permite entrenar el modelo, pintar las metricas en pantalla y devolver un diccionario de las metricas
        # lgbm.fit_and_get_metrics(save_metrics_dict=False)

        # ---- 2.3: [Metodo]: Metodo principal de la clase que ejecuta internamente 'fit_and_get_metrics', almacena el modelo junto con graficos e info adicional
        lgbm.fit_and_get_model_and_results()

        # ---- 2.4: [Metodo]: Permite ejecutar el analisis SHAP del modelo
        lgbm.execute_shap_analysis(
            sample=True,  # Realizar un sampleo de la data antes de calcular SHAP (!! Es muy intensivo en cómputo si no se samplea !!)
            num_features_to_show=100,  # En los gráficos se van a mostrar las N características mas importantes (con mas peso)
            num_sample=200,  # Número de instancias para background (explicador deep/grad).
        )

    def _xg_boost_classifier(self):
        """
        Metodo para realizar el entrenamiento del xgboost de clasificacion
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar
        # --------------------------------------------------------------------------------------

        # ---- 1.1: En este caso, los obtengo aleatorios (en el ejemplo de la dnn están todas las opciones)
        xgboost_hyperparams: Dict[str, Union[int, float]] = HxXgboostHyperparameterBuilder().get_random_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase que me va a proporcionar un toolkit para ese modelo en concreto
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados (Entrar en la clase para ver detalle)
        xgboost: HxXtremeGradientBoostingBinaryClassifier = HxXtremeGradientBoostingBinaryClassifier(
            data_dict=self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            hiperparams=xgboost_hyperparams,  # Diccionario con los hiperparametros del XGB
            metric="roc_auc",  # Literal['accuracy', 'precision', 'recall', 'f1', 'specificity', 'balanced_acc', 'roc_auc'], Metrica que vamos a evaluar
            problem_type="binary",  # Literal['binary', 'multiclass']
            save_path=None,  # Path donde se almacenarán los resultados (por defecto None, ya que se crean internamente las estructuras de carpetas)
            bins=None,  # En binary viene por defecto
            verbose="dont_show",  # Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show' (Cantidad de informacion adicional que se muestra en consola)
            random_state=42,  # Random state de incializacion de pesos para garantizar reproducibilidad

        )

        # ---- 2.2: [Metodo]: Permite entrenar el modelo, pintar las metricas en pantalla y devolver un diccionario de las metricas
        # xgboost.fit_and_get_metrics(save_metrics_dict=False)

        # ---- 2.3: [Metodo]: Metodo principal de la clase que ejecuta internamente 'fit_and_get_metrics', almacena el modelo junto con graficos e info adicional
        xgboost.fit_and_get_model_and_results()

        # ---- 2.4: [Metodo]: Permite ejecutar el analisis SHAP del modelo
        xgboost.execute_shap_analysis(
            sample=True,  # Realizar un sampleo de la data antes de calcular SHAP (!! Es muy intensivo en cómputo si no se samplea !!)
            num_features_to_show=100,  # En los gráficos se van a mostrar las N características mas importantes (con mas peso)
            num_sample=200,  # Número de instancias para background (explicador deep/grad).
        )

    def _cat_boost_classifier(self):
        """
        Metodo para realizar el entrenamiento del catboost de regresion
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar
        # --------------------------------------------------------------------------------------

        # ---- 1.1: En este caso, los obtengo aleatorios (en el ejemplo de la dnn están todas las opciones)
        catboost_hyperparams: Dict[str, Union[int, float]] = HxCatBoostHyperparameterBuilder().get_random_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase que me va a proporcionar un toolkit para ese modelo en concreto
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados (Entrar en la clase para ver detalle)
        catboost: HxCatBoostBinaryClassifier = HxCatBoostBinaryClassifier(
            data_dict=self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            hiperparams=catboost_hyperparams,  # Diccionario con los hiperparametros del XGB
            metric="roc_auc",  # Literal['accuracy', 'precision', 'recall', 'f1', 'specificity', 'balanced_acc', 'roc_auc'], Metrica que vamos a evaluar
            problem_type="binary",  # Literal['binary', 'multiclass']
            save_path=None,  # Path donde se almacenarán los resultados (por defecto None, ya que se crean internamente las estructuras de carpetas)
            bins=None,  # En binary viene por defecto
            verbose="dont_show",  # Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show' (Cantidad de informacion adicional que se muestra en consola)
            random_state=42,  # Random state de incializacion de pesos para garantizar reproducibilidad

        )

        # ---- 2.2: [Metodo]: Permite entrenar el modelo, pintar las metricas en pantalla y devolver un diccionario de las metricas
        # catboost.fit_and_get_metrics(save_metrics_dict=False)

        # ---- 2.3: [Metodo]: Metodo principal de la clase que ejecuta internamente 'fit_and_get_metrics', almacena el modelo junto con graficos e info adicional
        catboost.fit_and_get_model_and_results()

        # ---- 2.4: [Metodo]: Permite ejecutar el analisis SHAP del modelo
        catboost.execute_shap_analysis(
            sample=True,  # Realizar un sampleo de la data antes de calcular SHAP (!! Es muy intensivo en cómputo si no se samplea !!)
            num_features_to_show=100,  # En los gráficos se van a mostrar las N características mas importantes (con mas peso)
            num_sample=200,  # Número de instancias para background (explicador deep/grad).
        )

    # </editor-fold>






# ---- 1: ML Binary Classifiers
from hx_machine_learning_tools import HxLightGbmBinaryClassifier
from hx_machine_learning_tools import HxXtremeGradientBoostingBinaryClassifier
from hx_machine_learning_tools import HxCatBoostBinaryClassifier
from hx_machine_learning_tools import HxRandomForestBinaryClassifier

# ---- 2: ML Multiclass Classifiers
from hx_machine_learning_tools import HxLightGbmMulticlassClassifier
from hx_machine_learning_tools import HxXtremeGradientBoostingMulticlassClassifier
from hx_machine_learning_tools import HxCatBoostMulticlassClassifier

# ---- 3: ML Regressors
from hx_machine_learning_tools import HxLightGbmRegressor
from hx_machine_learning_tools import HxXtremeGradientBoostingRegressor
from hx_machine_learning_tools import HxCatBoostRegressor

# ---- 5: DL BinaryCLassifiers
from hx_deep_learning_tools import HxDenseNeuralNetworkBinaryClassifier

# ---- 6: DL MulticlassCLassifiers
from hx_deep_learning_tools import HxDenseNeuralNetworkMulticlassClassifier

# ---- 6: DL Regressors
from hx_deep_learning_tools import HxDenseNeuralNetworkRegressor